from dataclasses import dataclass, field
from typing import Optional, Union

import numpy as np
import numpy.typing as npt
from qibolab import AcquisitionType, AveragingMode, Parameter, Sweeper

from qibocal import update
from qibocal.auto.operation import Data, Parameters, QubitId, Results, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.config import log
from qibocal.protocols.utils import fallback_period, guess_period, readout_frequency
from qibocal.result import magnitude, phase

from . import utils


@dataclass
class RabiLengthSignalParameters(Parameters):
    """RabiLengthSignal runcard inputs."""

    pulse_duration_start: float
    """Initial pi pulse duration [ns]."""
    pulse_duration_end: float
    """Final pi pulse duration [ns]."""
    pulse_duration_step: float
    """Step pi pulse duration [ns]."""
    pulse_amplitude: Optional[float] = None
    """Pi pulse amplitude. Same for all qubits."""
    rx90: bool = False
    """Calibration of native pi pulse, if true calibrates pi/2 pulse"""
    interpolated_sweeper: bool = False
    """Use real-time interpolation if supported by instruments."""


@dataclass
class RabiLengthSignalResults(Results):
    """RabiLengthSignal outputs."""

    length: dict[QubitId, Union[int, list[float]]]
    """Pi pulse duration for each qubit."""
    amplitude: dict[QubitId, Union[float, list[float]]]
    """Pi pulse amplitude. Same for all qubits."""
    fitted_parameters: dict[QubitId, dict[str, float]]
    """Raw fitting output."""
    rx90: bool
    """Pi or Pi_half calibration"""


RabiLenSignalType = np.dtype(
    [("length", np.float64), ("signal", np.float64), ("phase", np.float64)]
)
"""Custom dtype for rabi amplitude."""


@dataclass
class RabiLengthSignalData(Data):
    """RabiLength acquisition outputs."""

    rx90: bool
    """Pi or Pi_half calibration"""
    amplitudes: dict[QubitId, float] = field(default_factory=dict)
    """Pulse durations provided by the user."""
    data: dict[QubitId, npt.NDArray[RabiLenSignalType]] = field(default_factory=dict)
    """Raw data acquired."""


def _acquisition(
    params: RabiLengthSignalParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> RabiLengthSignalData:
    r"""
    Data acquisition for RabiLength Experiment.
    In the Rabi experiment we apply a pulse at the frequency of the qubit and scan the drive pulse length
    to find the drive pulse length that creates a rotation of a desired angle.
    """

    sequence, qd_pulses, delays, ro_pulses, amplitudes = utils.sequence_length(
        targets, params, platform, params.rx90, use_align=params.interpolated_sweeper
    )
    sweep_range = (
        params.pulse_duration_start,
        params.pulse_duration_end,
        params.pulse_duration_step,
    )
    if params.interpolated_sweeper:
        sweeper = Sweeper(
            parameter=Parameter.duration_interpolated,
            range=sweep_range,
            pulses=[qd_pulses[q] for q in targets],
        )
    else:
        sweeper = Sweeper(
            parameter=Parameter.duration,
            range=sweep_range,
            pulses=[qd_pulses[q] for q in targets] + [delays[q] for q in targets],
        )

    data = RabiLengthSignalData(amplitudes=amplitudes, rx90=params.rx90)

    results = platform.execute(
        [sequence],
        [[sweeper]],
        updates=[
            {platform.qubits[q].probe: {"frequency": readout_frequency(q, platform)}}
            for q in targets
        ],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.INTEGRATION,
        averaging_mode=AveragingMode.CYCLIC,
    )

    for q in targets:
        result = results[ro_pulses[q].id]
        data.register_qubit(
            RabiLenSignalType,
            (q),
            dict(
                length=sweeper.values,
                signal=magnitude(result),
                phase=phase(result),
            ),
        )
    return data


def _fit(data: RabiLengthSignalData) -> RabiLengthSignalResults:
    """Post-processing for RabiLength experiment."""

    qubits = data.qubits
    fitted_parameters = {}
    durations = {}

    for qubit in qubits:
        qubit_data = data[qubit]
        rabi_parameter = qubit_data.length
        voltages = qubit_data.signal

        y_min = np.min(voltages)
        y_max = np.max(voltages)
        x_min = np.min(rabi_parameter)
        x_max = np.max(rabi_parameter)
        x = (rabi_parameter - x_min) / (x_max - x_min)
        y = (voltages - y_min) / (y_max - y_min) - 1 / 2

        period = fallback_period(guess_period(x, y))
        pguess = [0, np.sign(y[0]) * 0.5, period, 0, 0]
        try:
            popt, _, pi_pulse_parameter = utils.fit_length_function(
                x,
                y,
                pguess,
                signal=True,
                x_limits=(x_min, x_max),
                y_limits=(y_min, y_max),
            )
            durations[qubit] = pi_pulse_parameter
            fitted_parameters[qubit] = popt

        except Exception as e:
            log.warning(f"Rabi fit failed for qubit {qubit} due to {e}.")

    return RabiLengthSignalResults(
        durations, data.amplitudes, fitted_parameters, data.rx90
    )


def _update(
    results: RabiLengthSignalResults, platform: CalibrationPlatform, target: QubitId
):
    update.drive_duration(results.length[target], results.rx90, platform, target)
    update.drive_amplitude(results.amplitude[target], results.rx90, platform, target)


def _plot(data: RabiLengthSignalData, fit: RabiLengthSignalResults, target: QubitId):
    """Plotting function for RabiLength experiment."""
    return utils.plot(data, target, fit, data.rx90)


rabi_length_signal = Routine(_acquisition, _fit, _plot, _update)
"""RabiLength Routine object."""



from dataclasses import dataclass

from qibocal.auto.operation import QubitPairId, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.protocols.randomized_benchmarking.standard_rb import (
    StandardRBParameters,
    _plot,
)

from ...calibration.calibration import TwoQubitCalibration
from .utils import RB2QData, StandardRBResult, fit, twoq_rb_acquisition

FILE_CLIFFORDS = "2qubitCliffs.json"
FILE_INV = "2qubitCliffsInv.npz"


@dataclass
class StandardRB2QParameters(StandardRBParameters):
    """Parameters for the standard 2q randomized benchmarking protocol."""

    file: str = FILE_CLIFFORDS
    """File with the cliffords to be used."""
    file_inv: str = FILE_INV
    """File with the cliffords to be used in an inverted dict."""


def _acquisition(
    params: StandardRB2QParameters,
    platform: CalibrationPlatform,
    targets: list[QubitPairId],
) -> RB2QData:
    """Data acquisition for two qubit Standard Randomized Benchmarking."""

    return twoq_rb_acquisition(params, platform, targets)


def _fit(data: RB2QData) -> StandardRBResult:
    qubits = data.pairs
    results = fit(qubits, data)

    return results


def _update(
    results: StandardRBResult, platform: CalibrationPlatform, target: QubitPairId
):
    """Write rb fidelity in calibration."""
    if target not in platform.calibration.two_qubits:
        platform.calibration.two_qubits[target] = TwoQubitCalibration()

    platform.calibration.two_qubits[target].rb_fidelity = (
        results.fidelity[target],
        results.fit_uncertainties[target][1] / 2,
    )


standard_rb_2q = Routine(_acquisition, _fit, _plot, _update)



from dataclasses import dataclass, field

import numpy as np
import numpy.typing as npt
from qibolab import AcquisitionType, AveragingMode, Parameter, Sweeper

from qibocal import update
from qibocal.auto.operation import Data, QubitId, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.config import log
from qibocal.result import probability

from ..utils import chi2_reduced, fallback_period, guess_period
from . import utils
from .amplitude_signal import RabiAmplitudeSignalParameters, RabiAmplitudeSignalResults


@dataclass
class RabiAmplitudeParameters(RabiAmplitudeSignalParameters):
    """RabiAmplitude runcard inputs."""


@dataclass
class RabiAmplitudeResults(RabiAmplitudeSignalResults):
    """RabiAmplitude outputs."""

    chi2: dict[QubitId, list[float]] = field(default_factory=dict)


RabiAmpType = np.dtype(
    [("amp", np.float64), ("prob", np.float64), ("error", np.float64)]
)
"""Custom dtype for rabi amplitude."""


@dataclass
class RabiAmplitudeData(Data):
    """RabiAmplitude data acquisition."""

    rx90: bool
    """Pi or Pi_half calibration"""
    durations: dict[QubitId, float] = field(default_factory=dict)
    """Pulse durations provided by the user."""
    data: dict[QubitId, npt.NDArray[RabiAmpType]] = field(default_factory=dict)
    """Raw data acquired."""


def _acquisition(
    params: RabiAmplitudeParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> RabiAmplitudeData:
    r"""
    Data acquisition for Rabi experiment sweeping amplitude.
    In the Rabi experiment we apply a pulse at the frequency of the qubit and scan the drive pulse amplitude
    to find the drive pulse amplitude that creates a rotation of a desired angle.
    """

    sequence, qd_pulses, ro_pulses, durations = utils.sequence_amplitude(
        targets, params, platform, params.rx90
    )

    sweeper = Sweeper(
        parameter=Parameter.amplitude,
        range=(params.min_amp, params.max_amp, params.step_amp),
        pulses=[qd_pulses[qubit] for qubit in targets],
    )

    data = RabiAmplitudeData(durations=durations, rx90=params.rx90)

    # sweep the parameter
    results = platform.execute(
        [sequence],
        [[sweeper]],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.DISCRIMINATION,
        averaging_mode=AveragingMode.SINGLESHOT,
    )
    for qubit in targets:
        prob = probability(results[ro_pulses[qubit].id], state=1)
        data.register_qubit(
            RabiAmpType,
            (qubit),
            dict(
                amp=sweeper.values,
                prob=prob.tolist(),
                error=np.sqrt(prob * (1 - prob) / params.nshots).tolist(),
            ),
        )
    return data


def _fit(data: RabiAmplitudeData) -> RabiAmplitudeResults:
    """Post-processing for RabiAmplitude."""
    qubits = data.qubits

    pi_pulse_amplitudes = {}
    fitted_parameters = {}
    durations = {}
    chi2 = {}

    for qubit in qubits:
        qubit_data = data[qubit]

        x = qubit_data.amp
        y = qubit_data.prob

        period = fallback_period(guess_period(x, y))
        pguess = [0.5, 0.5, period, np.pi]
        try:
            popt, perr, pi_pulse_parameter = utils.fit_amplitude_function(
                x,
                y,
                pguess,
                sigma=qubit_data.error,
                signal=False,
            )
            pi_pulse_amplitudes[qubit] = [pi_pulse_parameter, perr[2] / 2]
            fitted_parameters[qubit] = popt.tolist()
            durations = {key: [value, 0] for key, value in data.durations.items()}
            chi2[qubit] = [
                chi2_reduced(
                    y,
                    utils.rabi_amplitude_function(x, *popt),
                    qubit_data.error,
                ),
                np.sqrt(2 / len(y)),
            ]

        except Exception as e:
            log.warning(f"Rabi fit failed for qubit {qubit} due to {e}.")
    return RabiAmplitudeResults(
        pi_pulse_amplitudes, durations, fitted_parameters, data.rx90, chi2
    )


def _plot(data: RabiAmplitudeData, target: QubitId, fit: RabiAmplitudeResults = None):
    """Plotting function for RabiAmplitude."""
    return utils.plot_probabilities(data, target, fit, data.rx90)


def _update(
    results: RabiAmplitudeResults, platform: CalibrationPlatform, target: QubitId
):
    update.drive_amplitude(results.amplitude[target], results.rx90, platform, target)
    update.drive_duration(results.length[target], results.rx90, platform, target)


rabi_amplitude = Routine(_acquisition, _fit, _plot, _update)
"""RabiAmplitude Routine object."""



from dataclasses import dataclass, field
from typing import Optional

import numpy as np
import numpy.typing as npt
from qibolab import (
    AcquisitionType,
    AveragingMode,
    Delay,
    Parameter,
    PulseSequence,
    Sweeper,
)
from scipy.optimize import curve_fit

from qibocal import update
from qibocal.auto.operation import QubitId, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.config import log

from ...result import magnitude, phase
from ...update import replace
from ..utils import (
    GHZ_TO_HZ,
    HZ_TO_GHZ,
    extract_feature,
    readout_frequency,
    table_dict,
    table_html,
)
from . import utils
from .qubit_flux_dependence import (
    QubitFluxData,
    QubitFluxParameters,
    QubitFluxResults,
    QubitFluxType,
)


@dataclass
class QubitCrosstalkParameters(QubitFluxParameters):
    """Crosstalk runcard inputs."""

    bias_point: Optional[dict[QubitId, float]] = field(default_factory=dict)
    """Dictionary with {qubit_id: bias_point_qubit_id}."""
    flux_qubits: Optional[list[QubitId]] = None
    """IDs of the qubits that we will sweep the flux on.
    If ``None`` flux will be swept on all qubits that we are running the routine on in a multiplex fashion.
    If given flux will be swept on the given qubits in a sequential fashion (n qubits will result to n different executions).
    Multiple qubits may be measured in each execution as specified by the ``qubits`` option in the runcard.
    """


@dataclass
class QubitCrosstalkData(QubitFluxData):
    """Crosstalk acquisition outputs when ``flux_qubits`` are given."""

    matrix_element: dict[QubitId, float] = field(default_factory=dict)
    """Diagonal flux element."""
    bias_point: dict[QubitId, float] = field(default_factory=dict)
    """Bias point for each qubit."""
    offset: dict[QubitId, float] = field(default_factory=dict)
    """Phase shift for each qubit."""
    qubit_frequency: dict[QubitId, float] = field(default_factory=dict)
    """Qubit frequency for each qubit."""
    data: dict[tuple[QubitId, QubitId], npt.NDArray[QubitFluxType]] = field(
        default_factory=dict
    )
    """Raw data acquired for (qubit, qubit_flux) pairs saved in nested dictionaries."""

    def register_qubit(self, qubit, flux_qubit, freq, bias, signal, phase):
        """Store output for single qubit."""
        ar = utils.create_data_array(freq, bias, signal, phase, dtype=QubitFluxType)
        if (qubit, flux_qubit) in self.data:
            self.data[qubit, flux_qubit] = np.rec.array(
                np.concatenate((self.data[qubit, flux_qubit], ar))
            )
        else:
            self.data[qubit, flux_qubit] = ar


@dataclass
class QubitCrosstalkResults(QubitFluxResults):
    """
    Qubit Crosstalk outputs.
    """

    qubit_frequency_bias_point: dict[QubitId, float] = field(default_factory=dict)
    """Expected qubit frequency at bias point."""
    crosstalk_matrix: dict[QubitId, dict[QubitId, float]] = field(default_factory=dict)
    """Crosstalk matrix element."""
    fitted_parameters: dict[tuple[QubitId, QubitId], dict] = field(default_factory=dict)
    """Fitted parameters for each couple target-flux qubit."""

    def __contains__(self, key: QubitId):
        """Checking if qubit is in crosstalk_matrix attribute."""
        return key in self.crosstalk_matrix


def _acquisition(
    params: QubitCrosstalkParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> QubitCrosstalkData:
    """Data acquisition for Crosstalk Experiment."""

    assert set(targets).isdisjoint(set(params.flux_qubits)), (
        "Flux qubits must be different from targets."
    )

    sequence = PulseSequence()
    ro_pulses = {}
    qd_pulses = {}
    offset = {}
    charging_energy = {}
    matrix_element = {}
    maximum_frequency = {}
    freq_sweepers = []
    offset_sweepers = []

    delta_frequency_range = np.arange(
        -params.freq_width / 2, params.freq_width / 2, params.freq_step
    )

    for qubit in targets:
        natives = platform.natives.single_qubit[qubit]
        charging_energy[qubit] = platform.calibration.single_qubits[
            qubit
        ].qubit.charging_energy
        qd_channel, qd_pulse = natives.RX()[0]
        ro_channel, ro_pulse = natives.MZ()[0]

        qd_pulse = replace(qd_pulse, duration=params.drive_duration)
        if params.drive_amplitude is not None:
            qd_pulse = replace(qd_pulse, amplitude=params.drive_amplitude)

        qd_pulses[qubit] = qd_pulse
        ro_pulses[qubit] = ro_pulse

        # store calibration parameters
        maximum_frequency[qubit] = platform.calibration.single_qubits[
            qubit
        ].qubit.maximum_frequency
        matrix_element[qubit] = platform.calibration.get_crosstalk_element(qubit, qubit)
        charging_energy[qubit] = platform.calibration.single_qubits[
            qubit
        ].qubit.charging_energy
        offset[qubit] = (
            -platform.calibration.single_qubits[qubit].qubit.sweetspot
            * matrix_element[qubit]
        )

        sequence.append((qd_channel, qd_pulse))
        sequence.append((ro_channel, Delay(duration=qd_pulse.duration)))
        sequence.append((ro_channel, ro_pulse))

        freq_sweepers.append(
            Sweeper(
                parameter=Parameter.frequency,
                values=platform.config(qd_channel).frequency + delta_frequency_range,
                channels=[qd_channel],
            )
        )

    delta_offset_range = np.arange(
        -params.bias_width / 2, params.bias_width / 2, params.bias_step
    )
    for q in params.flux_qubits:
        flux_channel = platform.qubits[q].flux
        offset0 = platform.config(flux_channel).offset
        offset_sweepers.append(
            Sweeper(
                parameter=Parameter.offset,
                values=offset0 + delta_offset_range,
                channels=[flux_channel],
            )
        )

    data = QubitCrosstalkData(
        resonator_type=platform.resonator_type,
        matrix_element=matrix_element,
        offset=offset,
        qubit_frequency=maximum_frequency,
        charging_energy=charging_energy,
        bias_point=params.bias_point,
    )

    options = dict(
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.INTEGRATION,
        averaging_mode=AveragingMode.CYCLIC,
    )

    updates = []
    for qubit in targets:
        if qubit in params.bias_point:
            channel = platform.qubits[qubit].flux
            updates.append({channel: {"offset": params.bias_point[qubit]}})

    updates += [
        {platform.qubits[q].probe: {"frequency": readout_frequency(q, platform)}}
        for q in targets
    ]

    for flux_qubit, offset_sweeper in zip(params.flux_qubits, offset_sweepers):
        results = platform.execute(
            [sequence], [[offset_sweeper], freq_sweepers], **options, updates=updates
        )

        # retrieve the results for every qubit
        for i, qubit in enumerate(targets):
            result = results[ro_pulses[qubit].id]
            data.register_qubit(
                qubit,
                flux_qubit,
                signal=magnitude(result),
                phase=phase(result),
                freq=freq_sweepers[i].values,
                bias=offset_sweeper.values,
            )
    return data


def _fit(data: QubitCrosstalkData) -> QubitCrosstalkResults:
    crosstalk_matrix = {qubit: {} for qubit in data.qubit_frequency}
    fitted_parameters = {}
    qubit_frequency_bias_point = {}

    for target_flux_qubit, qubit_data in data.data.items():
        frequencies, biases = extract_feature(
            qubit_data.freq,
            qubit_data.bias,
            qubit_data.signal,
            "max" if data.resonator_type == "2D" else "min",
        )
        target_qubit, flux_qubit = target_flux_qubit

        qubit_frequency_bias_point[target_qubit] = (
            utils.transmon_frequency(
                xi=data.bias_point[target_qubit],
                xj=0,
                d=0,
                w_max=data.qubit_frequency[target_qubit] * HZ_TO_GHZ,
                offset=data.offset[target_qubit],
                normalization=data.matrix_element[target_qubit],
                charging_energy=data.charging_energy[target_qubit] * HZ_TO_GHZ,
                crosstalk_element=1,
            )
            * GHZ_TO_HZ
        )

        def fit_function(x, crosstalk_element, offset):
            return utils.transmon_frequency(
                xi=data.bias_point[target_qubit],
                xj=x,
                d=0,
                w_max=data.qubit_frequency[target_qubit] * HZ_TO_GHZ,
                offset=offset,
                normalization=data.matrix_element[target_qubit],
                charging_energy=data.charging_energy[target_qubit] * HZ_TO_GHZ,
                crosstalk_element=crosstalk_element,
            )

        try:
            popt, _ = curve_fit(
                fit_function,
                biases,
                frequencies * HZ_TO_GHZ,
                bounds=((-np.inf, -1), (np.inf, 1)),
                maxfev=100000,
            )
            fitted_parameters[target_qubit, flux_qubit] = dict(
                xi=data.bias_point[target_qubit],
                d=0,
                w_max=data.qubit_frequency[target_qubit] * HZ_TO_GHZ,
                offset=popt[1],
                normalization=data.matrix_element[target_qubit],
                charging_energy=data.charging_energy[target_qubit] * HZ_TO_GHZ,
                crosstalk_element=float(popt[0]),
            )
            crosstalk_matrix[target_qubit][flux_qubit] = (
                popt[0] * data.matrix_element[target_qubit]
            )
        except RuntimeError as e:  # pragma: no cover
            log.error(
                f"Off-diagonal flux fit failed for qubit {flux_qubit} due to {e}."
            )

    return QubitCrosstalkResults(
        qubit_frequency_bias_point=qubit_frequency_bias_point,
        crosstalk_matrix=crosstalk_matrix,
        fitted_parameters=fitted_parameters,
    )


def _plot(data: QubitCrosstalkData, fit: QubitCrosstalkResults, target: QubitId):
    """Plotting function for Crosstalk Experiment."""
    figures, fitting_report = utils.flux_crosstalk_plot(
        data, target, fit, fit_function=utils.transmon_frequency
    )
    if fit is not None:
        labels = [
            "Qubit Frequency at Bias point [Hz]",
        ]
        values = [
            np.round(fit.qubit_frequency_bias_point[target], 4),
        ]
        for flux_qubit in fit.crosstalk_matrix[target]:
            if flux_qubit != target:
                labels.append(f"Crosstalk with qubit {flux_qubit} [V^-1]")
            else:
                labels.append("Flux dependence [V^-1]")
            values.append(np.round(fit.crosstalk_matrix[target][flux_qubit], 4))
        fitting_report = table_html(
            table_dict(
                target,
                labels,
                values,
            )
        )
    return figures, fitting_report


def _update(
    results: QubitCrosstalkResults, platform: CalibrationPlatform, qubit: QubitId
):
    """Update crosstalk matrix."""

    for flux_qubit, element in results.crosstalk_matrix[qubit].items():
        update.crosstalk_matrix(element, platform, qubit, flux_qubit)


qubit_crosstalk = Routine(_acquisition, _fit, _plot, _update)
"""Qubit crosstalk Routine object"""



from dataclasses import dataclass, field
from typing import Optional

import numpy as np
from qibolab import AcquisitionType, AveragingMode, Parameter, Sweeper

from qibocal.auto.operation import QubitId, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.result import probability

from . import t1
from .spin_echo_signal import SpinEchoSignalParameters, SpinEchoSignalResults, _update
from .utils import dynamical_decoupling_sequence, exponential_fit_probability, plot


@dataclass
class SpinEchoParameters(SpinEchoSignalParameters):
    """SpinEcho runcard inputs."""


@dataclass
class SpinEchoResults(SpinEchoSignalResults):
    """SpinEcho outputs."""

    chi2: Optional[dict[QubitId, tuple[float, Optional[float]]]] = field(
        default_factory=dict
    )
    """Chi squared estimate mean value and error."""


class SpinEchoData(t1.T1Data):
    """SpinEcho acquisition outputs."""


def _acquisition(
    params: SpinEchoParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> SpinEchoData:
    """Data acquisition for SpinEcho"""
    # create a sequence of pulses for the experiment:
    sequence, delays = dynamical_decoupling_sequence(platform, targets, kind="CP")

    # define the parameter to sweep and its range:
    # delay between pulses
    wait_range = np.arange(
        params.delay_between_pulses_start,
        params.delay_between_pulses_end,
        params.delay_between_pulses_step,
    )

    durations = []
    for q in targets:
        # this is assuming that RX and RX90 have the same duration
        duration = platform.natives.single_qubit[q].RX()[0][1].duration
        durations.append(duration)
        assert (params.delay_between_pulses_start - duration) / 2 >= 0, (
            f"Initial delay too short for qubit {q}, minimum delay should be {duration}"
        )
    assert len(set(durations)) == 1, (
        "Cannot run on mulitple qubit with different RX duration."
    )

    sweeper = Sweeper(
        parameter=Parameter.duration,
        values=(wait_range - durations[0]) / 2,
        pulses=delays,
    )

    results = platform.execute(
        [sequence],
        [[sweeper]],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.DISCRIMINATION,
        averaging_mode=AveragingMode.SINGLESHOT,
    )

    data = SpinEchoData()
    for qubit in targets:
        ro_pulse = list(sequence.channel(platform.qubits[qubit].acquisition))[-1]
        result = results[ro_pulse.id]
        prob = probability(result, state=1)
        error = np.sqrt(prob * (1 - prob) / params.nshots)
        data.register_qubit(
            t1.CoherenceProbType,
            (qubit),
            dict(
                wait=wait_range,
                prob=prob,
                error=error,
            ),
        )

    return data


def _fit(data: SpinEchoData) -> SpinEchoResults:
    """Post-processing for SpinEcho."""
    t2Echos, fitted_parameters, pcovs, chi2 = exponential_fit_probability(data)
    return SpinEchoResults(t2Echos, fitted_parameters, pcovs, chi2)


spin_echo = Routine(_acquisition, _fit, plot, _update)
"""SpinEcho Routine object."""



"""SWAP experiment for two qubit gates, chevron plot."""

from dataclasses import dataclass, field
from typing import Literal, Optional

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from qibolab import AcquisitionType, AveragingMode, Parameter, Sweeper
from scipy.optimize import curve_fit

from qibocal.auto.operation import Data, Parameters, QubitPairId, Results, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.config import log
from qibocal.protocols.utils import table_dict, table_html

from .... import update
from ..utils import fit_flux_amplitude, order_pair
from .utils import COLORAXIS, chevron_fit, chevron_sequence

__all__ = ["chevron"]


@dataclass
class ChevronParameters(Parameters):
    """CzFluxTime runcard inputs."""

    amplitude_min: float
    """Amplitude minimum."""
    amplitude_max: float
    """Amplitude maximum."""
    amplitude_step: float
    """Amplitude step."""
    duration_min: float
    """Duration minimum."""
    duration_max: float
    """Duration maximum."""
    duration_step: float
    """Duration step."""
    dt: Optional[int] = 0
    """Time delay between flux pulses and readout."""
    parking: bool = True
    """Wether to park non interacting qubits or not."""
    native: Literal["CZ", "iSWAP"] = "CZ"
    """Two qubit interaction to be calibrated."""


@dataclass
class ChevronResults(Results):
    """CzFluxTime outputs when fitting will be done."""

    amplitude: dict[QubitPairId, float]
    """CZ angle."""
    duration: dict[QubitPairId, int]
    """Virtual Z phase correction."""
    native: Literal["CZ", "iSWAP"] = "CZ"
    """Two qubit interaction to be calibrated."""


ChevronType = np.dtype(
    [
        ("amp", np.float64),
        ("length", np.float64),
        ("prob_high", np.float64),
        ("prob_low", np.float64),
    ]
)
"""Custom dtype for Chevron."""


@dataclass
class ChevronData(Data):
    """Chevron acquisition outputs."""

    native_amplitude: dict[QubitPairId, float] = field(default_factory=dict)
    """CZ platform amplitude for qubit pair."""
    native: str = "CZ"
    """Two qubit interaction to be calibrated.

    iSWAP and CZ are the possible options.

    """
    data: dict[QubitPairId, npt.NDArray[ChevronType]] = field(default_factory=dict)

    label: Optional[str] = None
    """Label for the data."""

    def register_qubit(self, low_qubit, high_qubit, length, amp, prob_low, prob_high):
        """Store output for single qubit."""
        size = len(length) * len(amp)
        amplitude, duration = np.meshgrid(amp, length)
        ar = np.empty(size, dtype=ChevronType)
        ar["length"] = duration.ravel()
        ar["amp"] = amplitude.ravel()
        ar["prob_low"] = prob_low.ravel()
        ar["prob_high"] = prob_high.ravel()
        self.data[low_qubit, high_qubit] = np.rec.array(ar)

    def amplitudes(self, pair):
        """Unique pair amplitudes."""
        return np.unique(self[pair].amp)

    def durations(self, pair):
        """Unique pair durations."""
        return np.unique(self[pair].length)

    def low_frequency(self, pair):
        return self[pair].prob_low

    def high_frequency(self, pair):
        return self[pair].prob_high


def _aquisition(
    params: ChevronParameters,
    platform: CalibrationPlatform,
    targets: list[QubitPairId],
) -> ChevronData:
    r"""Perform an CZ experiment between pairs of qubits by changing its
    frequency.

    Args:
        platform: CalibrationPlatform to use.
        params: Experiment parameters.
        targets (list): List of pairs to use sequentially.

    Returns:
        ChevronData: Acquisition data.
    """

    # create a DataUnits object to store the results
    data = ChevronData(native=params.native)
    for pair in targets:
        # order the qubits so that the low frequency one is the first
        ordered_pair = order_pair(pair, platform)
        sequence, flux_pulse, parking_pulses, delays = chevron_sequence(
            platform=platform,
            ordered_pair=ordered_pair,
            duration_max=params.duration_max,
            parking=params.parking,
            dt=params.dt,
            native=params.native,
        )
        sweeper_amplitude = Sweeper(
            parameter=Parameter.amplitude,
            range=(params.amplitude_min, params.amplitude_max, params.amplitude_step),
            pulses=[flux_pulse],
        )
        sweeper_duration = Sweeper(
            parameter=Parameter.duration,
            range=(params.duration_min, params.duration_max, params.duration_step),
            pulses=[flux_pulse] + delays + parking_pulses,
        )

        ro_high = list(sequence.channel(platform.qubits[ordered_pair[1]].acquisition))[
            -1
        ]
        ro_low = list(sequence.channel(platform.qubits[ordered_pair[0]].acquisition))[
            -1
        ]

        data.native_amplitude[ordered_pair] = flux_pulse.amplitude

        results = platform.execute(
            [sequence],
            [[sweeper_duration], [sweeper_amplitude]],
            nshots=params.nshots,
            relaxation_time=params.relaxation_time,
            acquisition_type=AcquisitionType.DISCRIMINATION,
            averaging_mode=AveragingMode.CYCLIC,
        )

        data.register_qubit(
            ordered_pair[0],
            ordered_pair[1],
            sweeper_duration.values,
            sweeper_amplitude.values,
            results[ro_low.id],
            results[ro_high.id],
        )
    return data


def _fit(data: ChevronData) -> ChevronResults:
    durations = {}
    amplitudes = {}
    for pair in data.data:
        amps = data.amplitudes(pair)
        times = data.durations(pair)

        signal = data.low_frequency(pair)
        signal_matrix = signal.reshape(len(times), len(amps)).T

        # guess amplitude computing FFT
        amplitude, index, delta = fit_flux_amplitude(signal_matrix, amps, times)
        # estimate duration by rabi curve at amplitude previously estimated
        y = signal_matrix[index, :].ravel()
        try:
            popt, _ = curve_fit(
                chevron_fit,
                times,
                y,
                p0=[delta * 2 * np.pi, np.pi, np.mean(y), np.mean(y)],
                bounds=(
                    [0, -2 * np.pi, np.min(y), np.min(y)],
                    [np.inf, 2 * np.pi, np.max(y), np.max(y)],
                ),
            )
            # duration can be estimated as the period of the oscillation
            duration = 1 / (popt[0] / 2 / np.pi)
            amplitudes[pair] = amplitude
            durations[pair] = int(duration)
        except Exception as e:
            log.warning(f"Chevron fit failed for pair {pair} due to {e}")

    return ChevronResults(amplitude=amplitudes, duration=durations, native=data.native)


def _plot(data: ChevronData, fit: ChevronResults, target: QubitPairId):
    """Plot the experiment result for a single pair."""
    # reverse qubit order if not found in data
    if target not in data.data:
        target = (target[1], target[0])

    pair_data = data[target]
    fig = make_subplots(
        rows=1,
        cols=2,
        subplot_titles=(
            f"Qubit {target[0]} - Low Frequency",
            f"Qubit {target[1]} - High Frequency",
        ),
    )
    fitting_report = ""

    fig.add_trace(
        go.Heatmap(
            x=pair_data.length,
            y=pair_data.amp,
            z=data.low_frequency(target),
            coloraxis=COLORAXIS[0],
        ),
        row=1,
        col=1,
    )

    fig.add_trace(
        go.Heatmap(
            x=pair_data.length,
            y=pair_data.amp,
            z=data.high_frequency(target),
            coloraxis=COLORAXIS[1],
        ),
        row=1,
        col=2,
    )

    for measured_qubit in target:
        if fit is not None:
            fig.add_trace(
                go.Scatter(
                    x=[
                        fit.duration[target],
                    ],
                    y=[
                        fit.amplitude[target],
                    ],
                    mode="markers",
                    marker=dict(
                        size=8,
                        color="black",
                        symbol="cross",
                    ),
                    name=f"{data.native} estimate",  #  Change name from the params
                    showlegend=True if measured_qubit == target[0] else False,
                    legendgroup="Voltage",
                ),
                row=1,
                col=1,
            )

    fig.update_layout(
        xaxis_title="Duration [ns]",
        xaxis2_title="Duration [ns]",
        yaxis_title=data.label or "Amplitude [a.u.]",
        legend=dict(orientation="h"),
    )
    fig.update_layout(
        coloraxis={"colorscale": "Oryel", "colorbar": {"x": 1.15}},
        coloraxis2={"colorscale": "Darkmint", "colorbar": {"x": -0.15}},
    )

    if fit is not None:
        fitting_report = table_html(
            table_dict(
                target[1],
                [f"{fit.native} amplitude", f"{fit.native} duration"],
                [
                    fit.amplitude[target],
                    fit.duration[target],
                ],
            )
        )

    return [fig], fitting_report


def _update(
    results: ChevronResults, platform: CalibrationPlatform, target: QubitPairId
):
    target = target[::-1] if target not in results.duration else target

    getattr(update, f"{results.native}_duration")(
        results.duration[target], platform, target
    )
    getattr(update, f"{results.native}_amplitude")(
        results.amplitude[target], platform, target
    )


chevron = Routine(_aquisition, _fit, _plot, _update, two_qubit_gates=True)
"""Chevron routine."""



from dataclasses import dataclass, field
from typing import Optional

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
from qibolab import AcquisitionType, AveragingMode, Parameter, Sweeper

from qibocal.auto.operation import Data, QubitId, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.result import probability

from ..utils import table_dict, table_html
from . import t1_signal, utils


@dataclass
class T1Parameters(t1_signal.T1SignalParameters):
    """T1 runcard inputs."""


@dataclass
class T1Results(t1_signal.T1SignalResults):
    """T1 outputs."""

    chi2: Optional[dict[QubitId, list[float]]] = field(default_factory=dict)
    """Chi squared estimate mean value and error."""


CoherenceProbType = np.dtype(
    [("wait", np.float64), ("prob", np.float64), ("error", np.float64)]
)
"""Custom dtype for coherence routines."""


@dataclass
class T1Data(Data):
    """T1 acquisition outputs."""

    data: dict[QubitId, npt.NDArray] = field(default_factory=dict)
    """Raw data acquired."""


def _acquisition(
    params: T1Parameters, platform: CalibrationPlatform, targets: list[QubitId]
) -> T1Data:
    """Data acquisition for T1 experiment."""

    sequence, ro_pulses, delays = t1_signal.t1_sequence(
        platform=platform, targets=targets
    )

    ro_wait_range = np.arange(
        params.delay_before_readout_start,
        params.delay_before_readout_end,
        params.delay_before_readout_step,
    )

    sweeper = Sweeper(
        parameter=Parameter.duration,
        values=ro_wait_range,
        pulses=[delays[q] for q in targets],
    )

    data = T1Data()

    results = platform.execute(
        [sequence],
        [[sweeper]],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.DISCRIMINATION,
        averaging_mode=AveragingMode.SINGLESHOT,
    )

    for qubit in targets:
        probs = probability(results[ro_pulses[qubit].id], state=1)
        errors = np.sqrt(probs * (1 - probs) / params.nshots)
        data.register_qubit(
            CoherenceProbType,
            (qubit),
            dict(wait=ro_wait_range, prob=probs, error=errors),
        )

    return data


def _fit(data: T1Data) -> T1Results:
    """
    Fitting routine for T1 experiment. The used model is

        .. math::

            y = p_0-p_1 e^{-x p_2}.
    """
    t1s, fitted_parameters, pcovs, chi2 = utils.exponential_fit_probability(data)
    return T1Results(t1s, fitted_parameters, pcovs, chi2)


def _plot(data: T1Data, target: QubitId, fit: T1Results = None):
    """Plotting function for T1 experiment."""

    figures = []
    fitting_report = ""
    qubit_data = data[target]
    waits = qubit_data.wait
    probs = qubit_data.prob
    error_bars = qubit_data.error

    fig = go.Figure(
        [
            go.Scatter(
                x=waits,
                y=probs,
                opacity=1,
                name="Probability of 1",
                showlegend=True,
                legendgroup="Probability of 1",
                mode="lines",
            ),
            go.Scatter(
                x=np.concatenate((waits, waits[::-1])),
                y=np.concatenate((probs + error_bars, (probs - error_bars)[::-1])),
                fill="toself",
                fillcolor=utils.COLORBAND,
                line=dict(color=utils.COLORBAND_LINE),
                showlegend=True,
                name="Errors",
            ),
        ]
    )

    if fit is not None:
        waitrange = np.linspace(
            min(waits),
            max(waits),
            2 * len(qubit_data),
        )

        params = fit.fitted_parameters[target]
        fig.add_trace(
            go.Scatter(
                x=waitrange,
                y=utils.exp_decay(waitrange, *params),
                name="Fit",
                line=go.scatter.Line(dash="dot"),
            )
        )
        fitting_report = table_html(
            table_dict(
                target,
                [
                    "T1 [ns]",
                    "chi2 reduced",
                ],
                [fit.t1[target], fit.chi2[target]],
                display_error=True,
            )
        )
    # last part
    fig.update_layout(
        showlegend=True,
        xaxis_title="Time [ns]",
        yaxis_title="Probability of State 1",
    )

    figures.append(fig)

    return figures, fitting_report


t1 = Routine(_acquisition, _fit, _plot, t1_signal._update)
"""T1 Routine object."""



from dataclasses import dataclass, field
from typing import Optional, Tuple

import numpy as np
import numpy.typing as npt
import plotly.express as px
from qibo import gates
from qibo.backends import construct_backend
from qibo.models import Circuit
from scipy.sparse import lil_matrix

from qibocal.auto.operation import Data, Parameters, QubitId, Results, Routine
from qibocal.auto.transpile import dummy_transpiler, execute_transpiled_circuit
from qibocal.calibration import CalibrationPlatform
from qibocal.config import log


@dataclass
class ReadoutMitigationMatrixParameters(Parameters):
    """ReadoutMitigationMatrix matrix inputs."""

    nshots: Optional[int] = None
    """Number of shots."""
    relaxation_time: Optional[int] = None
    """Relaxation time [ns]."""


@dataclass
class ReadoutMitigationMatrixResults(Results):
    readout_mitigation_matrix: dict[tuple[QubitId, ...], npt.NDArray[np.float64]] = (
        field(default_factory=dict)
    )
    """Readout mitigation matrices (inverse of measurement matrix)."""


ReadoutMitigationMatrixType = np.dtype(
    [
        ("state", int),
        ("frequency", np.float64),
    ]
)


ReadoutMitigationMatrixId = tuple[Tuple[QubitId, ...], str, str]
"""Data identifier for single list of qubits.

Tuple[QubitId, ...] is the qubits which have been passed on as parameters.
The two strings represents the expected state and the measured state.
"""


@dataclass
class ReadoutMitigationMatrixData(Data):
    """ReadoutMitigationMatrix acquisition outputs."""

    qubit_list: list[QubitId]
    """List of qubit ids"""
    nshots: int
    """Number of shots"""
    data: dict[ReadoutMitigationMatrixId, float] = field(default_factory=dict)
    """Raw data acquited."""


def _acquisition(
    params: ReadoutMitigationMatrixParameters,
    platform: CalibrationPlatform,
    targets: list[list[QubitId]],
) -> ReadoutMitigationMatrixData:
    data = ReadoutMitigationMatrixData(
        nshots=params.nshots, qubit_list=[list(qq) for qq in targets]
    )
    backend = construct_backend("qibolab", platform=platform)
    transpiler = dummy_transpiler(backend)

    for qubits in targets:
        nqubits = len(qubits)
        for i in range(2**nqubits):
            state = format(i, f"0{nqubits}b")
            c = Circuit(
                nqubits,
            )
            for q, bit in enumerate(state):
                if bit == "1":
                    c.add(gates.X(q))
            c.add(gates.M(*range(nqubits)))
            _, results = execute_transpiled_circuit(
                c, qubits, backend, nshots=params.nshots, transpiler=transpiler
            )
            frequencies = np.zeros(2 ** len(qubits))
            for i, freq in results.frequencies().items():
                frequencies[int(i, 2)] = freq
            for freq in frequencies:
                data.register_qubit(
                    ReadoutMitigationMatrixType,
                    (tuple(qubits)),
                    dict(
                        state=np.array([int(state, 2)]),
                        frequency=freq,
                    ),
                )
    return data


def _fit(data: ReadoutMitigationMatrixData) -> ReadoutMitigationMatrixResults:
    """Post processing for readout mitigation matrix protocol."""
    readout_mitigation_matrix = {}
    for qubits in data.qubit_list:
        qubit_data = data.data[tuple(qubits)]
        mitigation_matrix = []
        for state in range(2 ** len(qubits)):
            mitigation_matrix.append(qubit_data[qubit_data.state == state].frequency)
        mitigation_matrix = np.vstack(mitigation_matrix) / data.nshots
        try:
            readout_mitigation_matrix[tuple(qubits)] = np.linalg.inv(
                mitigation_matrix
            ).tolist()
        except np.linalg.LinAlgError as e:
            log.warning(f"ReadoutMitigationMatrix: the fitting was not succesful. {e}")
    res = ReadoutMitigationMatrixResults(
        readout_mitigation_matrix=readout_mitigation_matrix,
    )

    return res


def _plot(
    data: ReadoutMitigationMatrixData,
    fit: ReadoutMitigationMatrixResults,
    target: list[QubitId],
):
    """Plotting function for readout mitigation matrix."""
    fitting_report = ""
    figs = []
    if fit is not None:
        if tuple(target) in fit.readout_mitigation_matrix:
            computational_basis = [
                format(i, f"0{len(target)}b") for i in range(2 ** len(target))
            ]
            # use pinv since it should be already invertibile
            # however when casting to list we could lose precision
            measurement_matrix = np.linalg.pinv(
                fit.readout_mitigation_matrix[tuple(target)]
            )
            z = measurement_matrix
            fig = px.imshow(
                z,
                x=computational_basis,
                y=computational_basis,
                text_auto=True,
                labels={
                    "x": "Prepeared States",
                    "y": "Measured States",
                    "color": "Probabilities",
                },
                width=700,
                height=700,
            )
            figs.append(fig)
    return figs, fitting_report


def _update(
    results: ReadoutMitigationMatrixData,
    platform: CalibrationPlatform,
    target: list[QubitId],
):
    # create empty matrix if it doesn't exist
    if platform.calibration.readout_mitigation_matrix is None:
        platform.calibration.readout_mitigation_matrix = lil_matrix(
            (2**platform.calibration.nqubits, 2**platform.calibration.nqubits)
        )

    # compute indices
    mask = sum(1 << platform.calibration.qubit_index(i) for i in target)
    indices = [i for i in range(2**platform.calibration.nqubits) if (i & mask) == i]

    # update matrix
    platform.calibration.readout_mitigation_matrix[np.ix_(indices, indices)] = (
        results.readout_mitigation_matrix[tuple(target)]
    )


readout_mitigation_matrix = Routine(_acquisition, _fit, _plot, _update)
"""Readout mitigation matrix protocol."""



import io
import pathlib
from typing import Optional, Union

import plotly.graph_objects as go
from jinja2 import Environment, FileSystemLoader

from qibocal.auto.history import History
from qibocal.auto.operation import QubitId, QubitPairId
from qibocal.auto.output import Output
from qibocal.auto.task import Completed
from qibocal.config import log
from qibocal.web.report import (
    SCRIPT,
    STYLES,
    TEMPLATES,
    Report,
    report_css_styles,
    report_script,
)

ReportOutcome = tuple[str, list[go.Figure]]
"""Report produced by protocol."""


def generate_figures_and_report(
    node: Completed, target: Union[QubitId, QubitPairId, list[QubitId]]
) -> ReportOutcome:
    """Calling protocol plot by checking if fit has been performed.

    It operates on a completed `node` and a specific protocol `target`, generating
    a report outcome (cf. `ReportOutcome`).
    """
    if node.results is None:
        # plot acquisition data
        return node.task.operation.report(data=node.data, fit=None, target=target)
    if target not in node.results:
        # plot acquisition data and message for unsuccessful fit
        figures = node.task.operation.report(data=node.data, fit=None, target=target)[0]
        return figures, "An error occurred when performing the fit."
    # plot acquisition and fit
    return node.task.operation.report(data=node.data, fit=node.results, target=target)


def plotter(
    node: Completed, target: Union[QubitId, QubitPairId, list[QubitId]]
) -> tuple[str, str]:
    """Run plotly pipeline for generating html.

    Performs conversions of plotly figures in html rendered code for
    completed node on specific target.
    """
    figures, fitting_report = generate_figures_and_report(node, target)
    buffer = io.StringIO()
    html_list = []
    for figure in figures:
        figure.write_html(buffer, include_plotlyjs=False, full_html=False)
        buffer.seek(0)
        html_list.append(buffer.read())
    buffer.close()
    all_html = "".join(html_list)
    return all_html, fitting_report


def report(path: pathlib.Path, history: Optional[History] = None):
    """Report generation.

    Generates the report for protocol dumped in `path`.
    Executor can be passed to generate report on the fly.
    """

    if (path / "index.html").exists():  # pragma: no cover
        log.warning(f"Regenerating {path}/index.html")
    # load meta
    output = Output.load(path)

    if history is None:
        history = output.history

    env = Environment(loader=FileSystemLoader(TEMPLATES))
    template = env.get_template("template.html")
    html = template.render(
        is_static=True,
        css_styles=report_css_styles(STYLES),
        js_script=report_script(SCRIPT),
        path=path,
        title=path.name,
        report=Report(
            path=path,
            history=history,
            meta=output.meta.dump(),
            plotter=plotter,
        ),
    )

    (path / "index.html").write_text(html)



from dataclasses import dataclass

import numpy as np
import plotly.graph_objects as go
from qibolab import AcquisitionType, AveragingMode, Delay, Drag, Pulse, PulseSequence
from scipy.optimize import curve_fit

from qibocal import update
from qibocal.auto.operation import QubitId, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.protocols import drag
from qibocal.result import probability
from qibocal.update import replace

from .utils import COLORBAND, COLORBAND_LINE, table_dict, table_html

SEQUENCES = ["YpX9", "XpY9"]
"""Sequences used to fit drag parameter."""


@dataclass
class DragTuningSimpleParameters(drag.DragTuningParameters):
    """DragTuningSimple runcard inputs."""


@dataclass
class DragTuningSimpleResults(drag.DragTuningResults):
    """DragTuningSimple outputs."""

    def __contains__(self, key):
        return key in self.betas


@dataclass
class DragTuningSimpleData(drag.DragTuningData):
    """DragTuningSimple acquisition outputs."""

    def __getitem__(self, key: tuple[QubitId, str]):
        qubit, setup = key
        if setup == "YpX9":
            return self.data[qubit][::2]
        return self.data[qubit][1::2]


def add_drag(pulse: Pulse, beta: float) -> Pulse:
    """Add DRAG component to Gaussian Pulse."""
    return replace(pulse, envelope=Drag(rel_sigma=pulse.envelope.rel_sigma, beta=beta))


def _acquisition(
    params: DragTuningSimpleParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> DragTuningSimpleData:
    """Acquisition function for DRAG experiments.

    We execute two sequences YpX9 and XpY9 following
    https://rsl.yale.edu/sites/default/files/2024-08/2011-RSL-Thesis-Matthew-Reed.pdf
    for different value of the DRAG parameter.
    """

    data = DragTuningSimpleData()
    beta_range = np.arange(params.beta_start, params.beta_end, params.beta_step)

    sequences, all_ro_pulses = [], []
    for beta in beta_range:
        for setup in SEQUENCES:
            sequence = PulseSequence()
            ro_pulses = {}
            for q in targets:
                natives = platform.natives.single_qubit[q]
                ro_channel, ro_pulse = natives.MZ()[0]
                if setup == "YpX9":
                    qd_channel, ry = natives.R(phi=np.pi / 2)[0]
                    _, rx90 = natives.R(theta=np.pi / 2)[0]
                    sequence.append((qd_channel, add_drag(ry, beta=beta)))
                    sequence.append((qd_channel, add_drag(rx90, beta=beta)))
                    sequence.append(
                        (ro_channel, Delay(duration=rx90.duration + ry.duration))
                    )
                else:
                    qd_channel, rx = natives.RX()[0]
                    _, ry90 = natives.R(theta=np.pi / 2, phi=np.pi / 2)[0]
                    sequence.append((qd_channel, add_drag(rx, beta=beta)))
                    sequence.append((qd_channel, add_drag(ry90, beta=beta)))
                    sequence.append(
                        (ro_channel, Delay(duration=rx.duration + ry90.duration))
                    )
                sequence.append((ro_channel, ro_pulse))

            sequences.append(sequence)
            all_ro_pulses.append(
                {
                    qubit: list(sequence.channel(platform.qubits[q].acquisition))[-1]
                    for qubit in targets
                }
            )

    options = {
        "nshots": params.nshots,
        "relaxation_time": params.relaxation_time,
        "acquisition_type": AcquisitionType.DISCRIMINATION,
        "averaging_mode": AveragingMode.SINGLESHOT,
    }

    # execute the pulse sequence
    if params.unrolling:
        results = platform.execute(sequences, **options)
        for beta, ro_pulses in zip(np.repeat(beta_range, 2), all_ro_pulses):
            for qubit in targets:
                result = results[ro_pulses[qubit].id]
                prob = probability(result, state=1)
                # store the results
                data.register_qubit(
                    drag.DragTuningType,
                    (qubit),
                    dict(
                        prob=np.array([prob]),
                        error=np.array([np.sqrt(prob * (1 - prob) / params.nshots)]),
                        beta=np.array([beta]),
                    ),
                )
    else:
        for i, sequence in enumerate(sequences):
            result = platform.execute([sequence], **options)
            setup = "YpX9" if i % 2 == 2 else "XpY9"
            for qubit in targets:
                ro_pulse = list(sequence.channel(platform.qubits[qubit].acquisition))[
                    -1
                ]
                prob = probability(result[ro_pulse.id], state=1)
                # store the results
                data.register_qubit(
                    drag.DragTuningType,
                    (qubit),
                    dict(
                        prob=np.array([prob]),
                        error=np.array([np.sqrt(prob * (1 - prob) / params.nshots)]),
                        beta=np.array([beta_range[i // 2]]),
                    ),
                )

    return data


def _fit(data: DragTuningSimpleData) -> DragTuningSimpleResults:
    """Post-processing for DRAG protocol.

    A linear fit is applied for the probability of both sequences.
    The optimal is determined as the point in which the two lines met.
    """
    qubits = data.qubits
    fitted_parameters = {}
    betas_optimal = {}
    for qubit in qubits:
        for setup in SEQUENCES:
            qubit_data = data[qubit, setup]
            popt, _ = curve_fit(
                lambda x, a, b: a * x + b, qubit_data.beta, qubit_data.prob
            )
            fitted_parameters[qubit, setup] = popt.tolist()
        betas_optimal[qubit] = -(
            fitted_parameters[qubit, "YpX9"][1] - fitted_parameters[qubit, "XpY9"][1]
        ) / (fitted_parameters[qubit, "YpX9"][0] - fitted_parameters[qubit, "XpY9"][0])

    return DragTuningSimpleResults(betas_optimal, fitted_parameters)


def _plot(data: DragTuningSimpleData, target: QubitId, fit: DragTuningSimpleResults):
    """Plotting function for DragTuning."""

    figures = []
    fitting_report = ""

    fig = go.Figure()
    for setup in SEQUENCES:
        qubit_data = data[target, setup]
        fig.add_trace(
            go.Scatter(
                x=qubit_data.beta,
                y=qubit_data.prob,
                opacity=1,
                mode="lines",
                name=setup,
                showlegend=True,
                legendgroup=setup,
            )
        )

        fig.add_trace(
            go.Scatter(
                x=np.concatenate((qubit_data.beta, qubit_data.beta[::-1])),
                y=np.concatenate(
                    (
                        qubit_data.prob + qubit_data.error,
                        (qubit_data.prob - qubit_data.error)[::-1],
                    )
                ),
                fill="toself",
                fillcolor=COLORBAND,
                line=dict(color=COLORBAND_LINE),
                name=setup,
                showlegend=False,
                legendgroup=setup,
            )
        )

    # # add fitting traces
    if fit is not None:
        for setup in SEQUENCES:
            qubit_data = data[target, setup]
            betas = qubit_data.beta
            beta_range = np.linspace(
                min(betas),
                max(betas),
                20,
            )

            fig.add_trace(
                go.Scatter(
                    x=beta_range,
                    y=fit.fitted_parameters[target, setup][0] * betas
                    + fit.fitted_parameters[target, setup][1],
                    name=f"Fit {setup}",
                    line=go.scatter.Line(dash="dot"),
                ),
            )
        fitting_report = table_html(
            table_dict(
                target,
                ["Best DRAG parameter"],
                [np.round(fit.betas[target], 4)],
            )
        )

    fig.update_layout(
        showlegend=True,
        xaxis_title="Beta parameter",
        yaxis_title="Excited state probability",
    )

    figures.append(fig)

    return figures, fitting_report


def _update(
    results: DragTuningSimpleResults, platform: CalibrationPlatform, target: QubitId
):
    update.drag_pulse_beta(
        results.betas[target],
        platform,
        target,
    )


drag_simple = Routine(_acquisition, _fit, _plot, _update)
"""DragTuning Routine object."""



from dataclasses import dataclass

from qibolab import (
    AcquisitionType,
    AveragingMode,
    Delay,
    Parameter,
    PulseSequence,
    Sweeper,
)

from qibocal.auto.operation import QubitId, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.update import replace

from ... import update
from ...result import magnitude, phase
from ..utils import readout_frequency
from . import amplitude_signal, utils


@dataclass
class RabiAmplitudeEFParameters(amplitude_signal.RabiAmplitudeSignalParameters):
    """RabiAmplitudeEF runcard inputs."""


@dataclass
class RabiAmplitudeEFResults(amplitude_signal.RabiAmplitudeSignalResults):
    """RabiAmplitudeEF outputs."""


@dataclass
class RabiAmplitudeEFData(amplitude_signal.RabiAmplitudeSignalData):
    """RabiAmplitude data acquisition."""


def _acquisition(
    params: RabiAmplitudeEFParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> RabiAmplitudeEFData:
    r"""
    Data acquisition for Rabi EF experiment sweeping amplitude.

    The rabi protocol is performed after exciting the qubit to state 1.
    This protocol allows to compute the amplitude of the RX12 pulse to excite
    the qubit to state 2 starting from state 1.

    """

    # create a sequence of pulses for the experiment
    sequence = PulseSequence()
    qd_pulses = {}
    ro_pulses = {}
    durations = {}
    for q in targets:
        natives = platform.natives.single_qubit[q]
        qd_channel, qd_pulse = natives.RX()[0]
        qd12_channel, qd12_pulse = natives.RX12()[0]
        ro_channel, ro_pulse = natives.MZ()[0]

        if params.pulse_length is not None:
            qd12_pulse = replace(qd_pulse, duration=params.pulse_length)

        durations[q] = qd12_pulse.duration
        qd_pulses[q] = qd12_pulse
        ro_pulses[q] = ro_pulse

        sequence.append((qd_channel, qd_pulse))
        sequence.append((qd12_channel, Delay(duration=qd_pulse.duration)))
        sequence.append((qd12_channel, qd12_pulse))
        sequence.append(
            (qd_channel, Delay(duration=qd_pulse.duration + qd12_pulse.duration))
        )
        sequence.append(
            (ro_channel, Delay(duration=qd_pulse.duration + qd12_pulse.duration))
        )
        sequence.append((ro_channel, ro_pulse))

    sweeper = Sweeper(
        parameter=Parameter.amplitude,
        range=(params.min_amp, params.max_amp, params.step_amp),
        pulses=[qd_pulses[qubit] for qubit in targets],
    )

    assert not params.rx90, "Rabi ef available only for RX pulses."

    data = RabiAmplitudeEFData(durations=durations, rx90=False)

    # sweep the parameter
    results = platform.execute(
        [sequence],
        [[sweeper]],
        updates=[
            {
                platform.qubits[q].probe: {
                    "frequency": readout_frequency(q, platform, state=1)
                }
            }
            for q in targets
        ],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.INTEGRATION,
        averaging_mode=AveragingMode.CYCLIC,
    )
    for qubit in targets:
        result = results[ro_pulses[qubit].id]
        data.register_qubit(
            amplitude_signal.RabiAmpSignalType,
            (qubit),
            dict(
                amp=sweeper.values,
                signal=magnitude(result),
                phase=phase(result),
            ),
        )
    return data


def _plot(
    data: RabiAmplitudeEFData, target: QubitId, fit: RabiAmplitudeEFResults = None
):
    """Plotting function for RabiAmplitude."""
    figures, report = utils.plot(data, target, fit, data.rx90)
    if report is not None:
        report = report.replace("Pi pulse", "Pi pulse 12")
    return figures, report


def _update(
    results: RabiAmplitudeEFResults, platform: CalibrationPlatform, target: QubitId
):
    """Update RX2 amplitude_signal"""
    update.drive_12_amplitude(results.amplitude[target], platform, target)
    update.drive_12_duration(results.length[target], platform, target)


rabi_amplitude_ef = Routine(_acquisition, amplitude_signal._fit, _plot, _update)
"""RabiAmplitudeEF Routine object."""



from pathlib import Path

from qibo.backends import construct_backend

from ..auto.history import History
from ..auto.mode import AUTOCALIBRATION
from ..auto.output import Metadata, Output
from ..auto.runcard import Runcard
from ..calibration import CalibrationPlatform
from .report import report


def protocols_execution(runcard: Runcard, folder: Path, force, update):
    """Autocalibration.

    Arguments:

     - RUNCARD: runcard with declarative inputs.
    """
    # rename for brevity
    backend = construct_backend(backend=runcard.backend, platform=runcard.platform)
    platform = CalibrationPlatform.from_platform(backend.platform)

    if platform is None:
        raise ValueError("Qibocal requires a Qibolab platform to run.")

    # generate output folder
    path = Output.mkdir(folder, force)

    # dump action runcard
    runcard.dump(path)

    # generate meta
    meta = Metadata.generate(path.name, backend)
    meta.targets = runcard.targets
    output = Output(History(), meta, platform)
    output.dump(path)

    # connect and initialize platform
    platform.connect()

    # run
    meta.start()
    history = runcard.run(
        output=path,
        platform=platform,
        mode=AUTOCALIBRATION,
        update=update,
    )
    meta.end()

    # TODO: implement iterative dump of report...

    # stop and disconnect platform
    platform.disconnect()

    # dump history, metadata, and updated platform
    output.history = history
    output.dump(path)

    report(path, history)



from typing import Optional

import numpy as np
from qibolab import Delay, Platform, PulseSequence
from scipy.optimize import curve_fit

from qibocal.auto.operation import QubitId
from qibocal.protocols.utils import GHZ_TO_HZ, fallback_period, guess_period

POPT_EXCEPTION = [0, 0, 0, 0, 1]
"""Fit parameters output to handle exceptions"""
PERR_EXCEPTION = [1] * 5
"""Fit errors to handle exceptions; their choice has no physical meaning
and is meant to avoid breaking the code."""
THRESHOLD = 0.5
"""Threshold parameters for find_peaks to guess
   frequency for sinusoidal fit."""


def ramsey_sequence(
    platform: Platform,
    targets: list[QubitId],
    wait: int = 0,
    target_qubit: Optional[QubitId] = None,
):
    """Pulse sequence used in Ramsey (detuned) experiments.

    The pulse sequence is the following:

    RX90 -- wait -- RX90 -- MZ
    """
    delays = []
    sequence = PulseSequence()
    for qubit in targets:
        natives = platform.natives.single_qubit[qubit]

        qd_channel, qd_pulse = natives.R(theta=np.pi / 2)[0]
        ro_channel, ro_pulse = natives.MZ()[0]

        qd_delay = Delay(duration=wait)
        ro_delay = Delay(duration=wait)

        sequence.extend(
            [
                (qd_channel, qd_pulse),
                (qd_channel, qd_delay),
                (qd_channel, qd_pulse),
                (ro_channel, Delay(duration=2 * qd_pulse.duration)),
                (ro_channel, ro_delay),
                (ro_channel, ro_pulse),
            ]
        )

        delays.extend([qd_delay, ro_delay])
        if target_qubit is not None:
            assert target_qubit not in targets, (
                f"Cannot run Ramsey experiment on qubit {target_qubit} if it is already in Ramsey sequence."
            )
            natives = platform.natives.single_qubit[target_qubit]
            sequence += natives.RX()

    return sequence, delays


def ramsey_fit(x, offset, amplitude, delta, phase, decay):
    """Dumped sinusoidal fit."""
    return offset + amplitude * np.sin(x * delta + phase) * np.exp(-x * decay)


def fitting(x: list, y: list, errors: list = None) -> list:
    """
    Given the inputs list `x` and outputs one `y`, this function fits the
    `ramsey_fit` function and returns a list with the fit parameters.
    """
    y_max = np.max(y)
    y_min = np.min(y)
    x_max = np.max(x)
    x_min = np.min(x)
    delta_y = y_max - y_min
    delta_x = x_max - x_min
    y = (y - y_min) / delta_y
    x = (x - x_min) / delta_x
    err = errors / delta_y if errors is not None else None

    period = fallback_period(guess_period(x, y))
    omega = 2 * np.pi / period
    p0 = [
        0.5,
        0.5,
        omega,
        0,
        1,
    ]
    popt, perr = curve_fit(
        ramsey_fit,
        x,
        y,
        p0=p0,
        maxfev=5000,
        bounds=(
            [0, 0, 0, -np.pi, 0],
            [1, 1, np.inf, np.pi, np.inf],
        ),
        sigma=err,
    )
    popt = [
        delta_y * popt[0] + y_min,
        delta_y * popt[1] * np.exp(x_min * popt[4] / delta_x),
        popt[2] / delta_x,
        popt[3] - x_min * popt[2] / delta_x,
        popt[4] / delta_x,
    ]
    perr = np.sqrt(np.diag(perr))
    perr = [
        delta_y * perr[0],
        delta_y
        * np.exp(x_min * popt[4] / delta_x)
        * np.sqrt(perr[1] ** 2 + (popt[1] * x_min * perr[4] / delta_x) ** 2),
        perr[2] / delta_x,
        np.sqrt(perr[3] ** 2 + (perr[2] * x_min / delta_x) ** 2),
        perr[4] / delta_x,
    ]
    return popt, perr


def process_fit(
    popt: list[float], perr: list[float], qubit_frequency: float, detuning: float
):
    """Processing Ramsey fitting results."""

    delta_fitting = popt[2] / (2 * np.pi)
    if detuning is not None:
        sign = np.sign(detuning)
        delta_phys = int(sign * (delta_fitting * GHZ_TO_HZ - np.abs(detuning)))
    else:
        delta_phys = int(delta_fitting * GHZ_TO_HZ)

    corrected_qubit_frequency = int(qubit_frequency - delta_phys)
    t2 = 1 / popt[4]
    new_frequency = [
        corrected_qubit_frequency,
        perr[2] * GHZ_TO_HZ / (2 * np.pi),
    ]
    t2 = [t2, perr[4] * (t2**2)]

    delta_phys_measure = [
        -delta_phys,
        perr[2] * GHZ_TO_HZ / (2 * np.pi),
    ]
    delta_fitting_measure = [
        -delta_fitting * GHZ_TO_HZ,
        perr[2] * GHZ_TO_HZ / (2 * np.pi),
    ]

    return new_frequency, t2, delta_phys_measure, delta_fitting_measure, popt



"""Helper functions to update parameters in platform."""

from collections.abc import Iterable
from typing import Union

import numpy as np
from pydantic import BaseModel
from qibolab import Platform, PulseSequence, VirtualZ

from qibocal.auto.operation import QubitId, QubitPairId

CLASSIFICATION_PARAMS = [
    "threshold",
    "iq_angle",
    "mean_gnd_states",
    "mean_exc_states",
    "classifier_hpars",
]


def replace(model: BaseModel, **update):
    """Replace interface for pydantic models."""
    return model.model_copy(update=update)


def readout_frequency(freq: float, platform: Platform, qubit: QubitId):
    """Update readout frequency value in platform for specific qubit."""
    ro_channel = platform.qubits[qubit].probe
    platform.update({f"configs.{ro_channel}.frequency": freq})


def bare_resonator_frequency(freq: float, platform: Platform, qubit: QubitId):
    """Update rbare frequency value in platform for specific qubit."""
    platform.calibration.single_qubits[qubit].resonator.bare_frequency = int(freq)


def dressed_resonator_frequency(freq: float, platform: Platform, qubit: QubitId):
    """Update rbare frequency value in platform for specific qubit."""
    platform.calibration.single_qubits[qubit].resonator.dressed_frequency = int(freq)


def readout_amplitude(amp: float, platform: Platform, qubit: QubitId):
    """Update readout amplitude value in platform for specific qubit."""
    platform.update({f"native_gates.single_qubit.{qubit}.MZ.0.1.probe.amplitude": amp})


def drive_frequency(
    freq: Union[float, tuple, list], platform: Platform, qubit: QubitId
):
    """Update drive frequency value in platform for specific qubit."""
    if isinstance(freq, Iterable):
        freq = freq[0]
    drive_channel = platform.qubits[qubit].drive
    platform.update({f"configs.{drive_channel}.frequency": freq})


def drive_amplitude(
    amp: Union[float, tuple, list], rx90: bool, platform: Platform, qubit: QubitId
):
    """Update drive frequency value in platform for specific qubit."""
    if isinstance(amp, Iterable):
        amp = amp[0]
    if rx90:
        platform.update({f"native_gates.single_qubit.{qubit}.RX90.0.1.amplitude": amp})
    else:
        platform.update({f"native_gates.single_qubit.{qubit}.RX.0.1.amplitude": amp})


def drive_duration(
    duration: Union[int, tuple, list], rx90: bool, platform: Platform, qubit: QubitId
):
    """Update drive duration value in platform for specific qubit."""
    if isinstance(duration, Iterable):
        duration = duration[0]
    if rx90:
        platform.update(
            {f"native_gates.single_qubit.{qubit}.RX90.0.1.duration": int(duration)}
        )
    else:
        platform.update(
            {f"native_gates.single_qubit.{qubit}.RX.0.1.duration": int(duration)}
        )


def crosstalk_matrix(
    matrix_element: float, platform: Platform, qubit: QubitId, flux_qubit: QubitId
):
    """Update crosstalk_matrix element."""
    if platform.calibration.flux_crosstalk_matrix is None:
        platform.calibration.flux_crosstalk_matrix = np.zeros(
            (platform.calibration.nqubits, platform.calibration.nqubits)
        )
    platform.calibration.set_crosstalk_element(qubit, flux_qubit, matrix_element)


def iq_angle(angle: float, platform: Platform, qubit: QubitId):
    """Update iq angle value in platform for specific qubit."""
    ro_channel = platform.qubits[qubit].acquisition
    platform.update({f"configs.{ro_channel}.iq_angle": angle})


def threshold(threshold: float, platform: Platform, qubit: QubitId):
    ro_channel = platform.qubits[qubit].acquisition
    platform.update({f"configs.{ro_channel}.threshold": threshold})


def mean_gnd_states(ground_state: list, platform: Platform, qubit: QubitId):
    """Update mean ground state value in platform for specific qubit."""
    platform.calibration.single_qubits[qubit].readout.ground_state = ground_state


def mean_exc_states(excited_state: list, platform: Platform, qubit: QubitId):
    """Update mean excited state value in platform for specific qubit."""
    platform.calibration.single_qubits[qubit].readout.excited_state = excited_state


def readout_fidelity(fidelity: float, platform: Platform, qubit: QubitId):
    """Update fidelity of single shot classification."""
    platform.calibration.single_qubits[qubit].readout.fidelity = float(fidelity)


def virtual_phases(
    phases: dict[QubitId, float], native: str, platform: Platform, pair: QubitPairId
):
    native_sequence = getattr(platform.natives.two_qubit[pair], native)
    new_native = PulseSequence()
    if len(native_sequence) > 1:
        new_native.append(native_sequence[0])
    else:  # pragma: no cover
        new_native = native_sequence
    for qubit, phase in phases.items():
        new_native.append((platform.qubits[qubit].drive, VirtualZ(phase=phase)))

    platform.update(
        {f"native_gates.two_qubit.{f'{pair[0]}-{pair[1]}'}.{native}": new_native}
    )


def CZ_duration(duration: int, platform: Platform, pair: QubitPairId):
    """Update CZ duration for specific pair."""
    platform.update(
        {f"native_gates.two_qubit.{f'{pair[0]}-{pair[1]}'}.CZ.0.1.duration": duration}
    )


def CZ_amplitude(amp: float, platform: Platform, pair: QubitPairId):
    """Update CZ amplitude for specific pair."""
    platform.update(
        {f"native_gates.two_qubit.{f'{pair[0]}-{pair[1]}'}.CZ.0.1.amplitude": amp}
    )


def iSWAP_duration(duration: int, platform: Platform, pair: QubitPairId):
    """Update iSWAP_duration duration for specific pair."""
    platform.update(
        {f"native_gates.two_qubit.{f'{pair[0]}-{pair[1]}'}.CZ.0.1.duration": duration}
    )


def iSWAP_amplitude(amp: float, platform: Platform, pair: QubitPairId):
    """Update iSWAP_duration amplitude for specific pair."""
    platform.update(
        {f"native_gates.two_qubit.{f'{pair[0]}-{pair[1]}'}.CZ.0.1.amplitude": amp}
    )


def t1(t1: int, platform: Platform, qubit: QubitId):
    """Update t1 value in platform for specific qubit."""
    platform.calibration.single_qubits[qubit].t1 = tuple(t1)


def t2(t2: int, platform: Platform, qubit: QubitId):
    """Update t2 value in platform for specific qubit."""
    platform.calibration.single_qubits[qubit].t2 = tuple(t2)


def t2_spin_echo(t2_spin_echo: float, platform: Platform, qubit: QubitId):
    """Update t2 echo value in platform for specific qubit."""
    platform.calibration.single_qubits[qubit].t2_spin_echo = tuple(t2_spin_echo)


def drag_pulse_beta(beta: float, platform: Platform, qubit: QubitId):
    """Update beta parameter value in platform for specific qubit."""
    platform.update(
        {
            f"native_gates.single_qubit.{qubit}.RX.0.1.envelope.kind": "drag",
            f"native_gates.single_qubit.{qubit}.RX.0.1.envelope.beta": beta,
        }
    )


def sweetspot(sweetspot: float, platform: Platform, qubit: QubitId):
    """Update sweetspot parameter in platform for specific qubit."""
    platform.calibration.single_qubits[qubit].qubit.sweetspot = float(sweetspot)


def flux_coefficients(
    flux_coefficients: list[float], platform: Platform, qubit: QubitId
):
    """Update flux-amplitude relation parameters for specific qubit."""
    platform.calibration.single_qubits[qubit].qubit.flux_coefficients = [
        float(value) for value in flux_coefficients
    ]


def flux_offset(offset: float, platform: Platform, qubit: QubitId):
    """Update flux offset parameter in platform for specific qubit."""
    platform.update({f"configs.{platform.qubits[qubit].flux}.offset": offset})


def frequency_12_transition(frequency: int, platform: Platform, qubit: QubitId):
    channel = platform.qubits[qubit].drive_qudits[1, 2]
    platform.update({f"configs.{channel}.frequency": frequency})
    platform.calibration.single_qubits[qubit].qubit.frequency_12 = int(frequency)


def drive_12_amplitude(amplitude: float, platform: Platform, qubit: QubitId):
    platform.update(
        {f"native_gates.single_qubit.{qubit}.RX12.0.1.amplitude": amplitude}
    )


def drive_12_duration(
    duration: Union[int, tuple, list], platform: Platform, qubit: QubitId
):
    """Update drive duration value in platform for specific qubit."""
    platform.update(
        {f"native_gates.single_qubit.{qubit}.RX12.0.1.duration": int(duration)}
    )


def coupling(g: float, platform: Platform, qubit: QubitId):
    platform.calibration.single_qubits[qubit].readout.coupling = float(g)


def kernel(kernel: np.ndarray, platform: Platform, qubit: QubitId):
    ro_channel = platform.qubits[qubit].acquisition
    platform.update({f"configs.{ro_channel}.kernel": kernel})


def feedback(feedback: list[float], platform: Platform, qubit: QubitId):
    """Update flux pulse feedback filter parameter in platform for specific qubit."""
    feedbackQM = feedback.copy()
    feedbackQM = [-feedbackQM[1]]
    platform.update(
        {f"configs.{platform.qubits[qubit].flux}.filter.feedback": feedbackQM}
    )


def feedforward(feedforward: list[float], platform: Platform, qubit: QubitId):
    """Update flux pulse feedforward parameter in platform for specific qubit."""
    platform.update(
        {f"configs.{platform.qubits[qubit].flux}.filter.feedforward": feedforward}
    )



from dataclasses import dataclass, field
from typing import Optional

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
from qibolab import AcquisitionType, AveragingMode, Parameter, Readout, Sweeper

from qibocal.calibration import CalibrationPlatform

from ...auto.operation import QubitId, Routine
from ...config import log
from ...result import probability
from ..utils import table_dict, table_html
from .ramsey import (
    COLORBAND,
    COLORBAND_LINE,
    RamseySignalData,
    RamseySignalParameters,
    RamseySignalResults,
    RamseyType,
    _update,
)
from .utils import fitting, process_fit, ramsey_fit, ramsey_sequence


@dataclass
class RamseyZZParameters(RamseySignalParameters):
    """RamseyZZ runcard inputs."""

    target_qubit: Optional[QubitId] = None
    """Target qubit that will be excited."""


@dataclass
class RamseyZZResults(RamseySignalResults):
    """RamseyZZ outputs."""

    zz: dict[QubitId, float] = field(default_factory=dict)
    coupling: dict[QubitId, float] = field(default_factory=dict)

    def __contains__(self, qubit: QubitId):
        return qubit in self.zz


@dataclass
class RamseyZZData(RamseySignalData):
    """RamseyZZ acquisition outputs."""

    target_qubit: Optional[QubitId] = None
    """Qubit that will be excited."""
    anharmonicity: dict[QubitId, float] = field(default_factory=dict)
    """Targets qubit anharmonicity."""
    anharmonicity_target_qubit: float = None
    """Anharmonicity of target qubit."""
    frequency_target_qubit: float = 0
    "Frequency of target qubit."
    data: dict[tuple[QubitId, str], npt.NDArray[RamseyType]] = field(
        default_factory=dict
    )
    """Raw data acquired."""


def _acquisition(
    params: RamseyZZParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> RamseyZZData:
    """Data acquisition for RamseyZZ Experiment.

    Standard Ramsey experiment repeated twice.
    In the second execution one qubit is brought to the excited state.
    """

    waits = np.arange(
        params.delay_between_pulses_start,
        params.delay_between_pulses_end,
        params.delay_between_pulses_step,
    )

    data = RamseyZZData(
        detuning=params.detuning,
        qubit_freqs={
            qubit: platform.config(platform.qubits[qubit].drive).frequency
            for qubit in targets
        },
        anharmonicity={
            qubit: platform.calibration.single_qubits[qubit].qubit.anharmonicity
            for qubit in targets
        },
        anharmonicity_target_qubit=platform.calibration.single_qubits[
            params.target_qubit
        ].qubit.anharmonicity,
        frequency_target_qubit=platform.config(
            platform.qubits[params.target_qubit].drive
        ).frequency,
        target_qubit=params.target_qubit,
    )

    updates = []
    if params.detuning is not None:
        for qubit in targets:
            channel = platform.qubits[qubit].drive
            f0 = platform.config(channel).frequency
            updates.append({channel: {"frequency": f0 + params.detuning}})

    for setup in ["I", "X"]:
        if not params.unrolling:
            sequence, delays = ramsey_sequence(
                platform=platform,
                targets=targets,
                target_qubit=params.target_qubit if setup == "X" else None,
            )

            sweeper = Sweeper(
                parameter=Parameter.duration,
                values=waits,
                pulses=delays,
            )

            # execute the sweep
            results = platform.execute(
                [sequence],
                [[sweeper]],
                nshots=params.nshots,
                relaxation_time=params.relaxation_time,
                acquisition_type=AcquisitionType.DISCRIMINATION,
                averaging_mode=AveragingMode.SINGLESHOT,
                updates=updates,
            )

            for qubit in targets:
                ro_pulse = list(sequence.channel(platform.qubits[qubit].acquisition))[
                    -1
                ]
                probs = probability(results[ro_pulse.id], state=1)
                errors = [np.sqrt(prob * (1 - prob) / params.nshots) for prob in probs]

        else:
            sequences, all_ro_pulses = [], []
            probs, errors = [], []
            for wait in waits:
                sequence, _ = ramsey_sequence(
                    platform=platform,
                    targets=targets,
                    wait=wait,
                    target_qubit=params.target_qubit if setup == "X" else None,
                )
                sequences.append(sequence)
                all_ro_pulses.append(
                    {
                        qubit: [
                            readout
                            for readout in sequence.channel(
                                platform.qubits[qubit].acquisition
                            )
                            if isinstance(readout, Readout)
                        ][0]
                        for qubit in targets
                    }
                )

            results = platform.execute(
                sequences,
                nshots=params.nshots,
                relaxation_time=params.relaxation_time,
                acquisition_type=AcquisitionType.DISCRIMINATION,
                averaging_mode=AveragingMode.SINGLESHOT,
                updates=updates,
            )

            for wait, ro_pulses in zip(waits, all_ro_pulses):
                for qubit in targets:
                    result = results[ro_pulses[qubit].id]
                    prob = probability(result, state=1)
                    probs.append(prob)
                    errors.append(np.sqrt(prob * (1 - prob) / params.nshots))

        for qubit in targets:
            data.register_qubit(
                RamseyType,
                (qubit, setup),
                dict(
                    wait=waits,
                    prob=probs,
                    errors=errors,
                ),
            )

    return data


def _fit(data: RamseyZZData) -> RamseyZZResults:
    """Fitting procedure for RamseyZZ protocol.

    Standard Ramsey fitting procedure is applied for both version of
    the experiment.

    """
    waits = data.waits
    popts = {}
    freq_measure = {}
    t2_measure = {}
    delta_phys_measure = {}
    delta_fitting_measure = {}
    zz = {}
    coupling = {}
    for qubit in data.qubits:
        for setup in ["I", "X"]:
            qubit_data = data[qubit, setup]
            qubit_freq = data.qubit_freqs[qubit]
            probs = qubit_data["prob"]
            try:
                popt, perr = fitting(waits, probs, qubit_data.errors)
                (
                    freq_measure[qubit, setup],
                    t2_measure[qubit, setup],
                    delta_phys_measure[qubit, setup],
                    delta_fitting_measure[qubit, setup],
                    popts[qubit, setup],
                ) = process_fit(popt, perr, qubit_freq, data.detuning)
            except Exception as e:
                log.warning(f"Ramsey fitting failed for qubit {qubit} due to {e}.")
        # compute zz and qq coupling
        # zz the difference in frequency between the two measurement
        zz[qubit] = float(
            np.abs(freq_measure[qubit, "X"][0] - freq_measure[qubit, "I"][0])
        )

        # coupling computing by inverting the following formula
        # xi = 2 g**2 (1 / (delta_q - alpha_1) - 1 / (delta_q + alpha_2))
        # where delta_q is the difference in frequency and alpha_i is the anharmonicity
        delta_qubit_freq = np.abs(data.qubit_freqs[qubit] - data.frequency_target_qubit)
        denominator = -(
            1 / (delta_qubit_freq - data.anharmonicity[qubit])
            - 1 / (delta_qubit_freq + data.anharmonicity_target_qubit)
        )
        coupling[qubit] = float(np.sqrt(zz[qubit] / 2 / np.abs(denominator)))
    return RamseyZZResults(
        detuning=data.detuning,
        frequency=freq_measure,
        t2=t2_measure,
        delta_phys=delta_phys_measure,
        delta_fitting=delta_fitting_measure,
        fitted_parameters=popts,
        zz=zz,
        coupling=coupling,
    )


def _plot(data: RamseyZZData, target: QubitId, fit: RamseyZZResults = None):
    """Plotting function for Ramsey Experiment."""
    figures = []
    fitting_report = ""

    waits = data[target, "I"].wait
    probs_I = data.data[target, "I"].prob
    probs_X = data.data[target, "X"].prob

    error_bars_I = data.data[target, "I"].errors
    error_bars_X = data.data[target, "X"].errors
    fig = go.Figure(
        [
            go.Scatter(
                x=waits,
                y=probs_I,
                opacity=1,
                name="I",
                showlegend=True,
                legendgroup="I ",
                mode="lines",
            ),
            go.Scatter(
                x=np.concatenate((waits, waits[::-1])),
                y=np.concatenate(
                    (probs_I + error_bars_I, (probs_I - error_bars_I)[::-1])
                ),
                fill="toself",
                fillcolor=COLORBAND,
                line=dict(color=COLORBAND_LINE),
                showlegend=True,
                name="Errors I",
            ),
            go.Scatter(
                x=waits,
                y=probs_X,
                opacity=1,
                name="X",
                showlegend=True,
                legendgroup="X",
                mode="lines",
            ),
            go.Scatter(
                x=np.concatenate((waits, waits[::-1])),
                y=np.concatenate(
                    (probs_X + error_bars_X, (probs_X - error_bars_X)[::-1])
                ),
                fill="toself",
                fillcolor=COLORBAND,
                line=dict(color=COLORBAND_LINE),
                showlegend=True,
                name="Errors X",
            ),
        ]
    )

    if fit is not None:
        fig.add_trace(
            go.Scatter(
                x=waits,
                y=ramsey_fit(waits, *fit.fitted_parameters[target, "I"]),
                name="Fit I",
                line=go.scatter.Line(dash="dot"),
            )
        )

        fig.add_trace(
            go.Scatter(
                x=waits,
                y=ramsey_fit(waits, *fit.fitted_parameters[target, "X"]),
                name="Fit X",
                line=go.scatter.Line(dash="dot"),
            )
        )
        fitting_report = table_html(
            table_dict(
                target,
                [
                    f"ZZ  with {data.target_qubit} [kHz]",
                    f"Coupling with {data.target_qubit} [MHz]",
                ],
                [
                    np.round(
                        fit.zz[target] * 1e-3,
                        0,
                    ),
                    np.round(
                        fit.coupling[target] * 1e-6,
                        2,
                    ),
                ],
            )
        )

    fig.update_layout(
        showlegend=True,
        xaxis_title="Time [ns]",
        yaxis_title="Excited state probability",
    )

    figures.append(fig)

    return figures, fitting_report


ramsey_zz = Routine(_acquisition, _fit, _plot, _update)
"""Ramsey Routine object."""



from dataclasses import dataclass, field
from typing import Optional

import numpy as np
from qibolab import AcquisitionType, AveragingMode, Parameter, Sweeper

from qibocal.auto.operation import QubitId, Routine
from qibocal.calibration import CalibrationPlatform

from ...result import probability
from ..ramsey.utils import ramsey_sequence
from . import t1, t2_signal, utils


@dataclass
class T2Parameters(t2_signal.T2SignalParameters):
    """T2 runcard inputs."""

    delay_between_pulses_start: int
    """Initial delay between RX(pi/2) pulses in ns."""
    delay_between_pulses_end: int
    """Final delay between RX(pi/2) pulses in ns."""
    delay_between_pulses_step: int
    """Step delay between RX(pi/2) pulses in ns."""


@dataclass
class T2Results(t2_signal.T2SignalResults):
    """T2 outputs."""

    chi2: Optional[dict[QubitId, tuple[float, Optional[float]]]] = field(
        default_factory=dict
    )
    """Chi squared estimate mean value and error."""


class T2Data(t1.T1Data):
    """T2 acquisition outputs."""


def _acquisition(
    params: T2Parameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> T2Data:
    """Data acquisition for T2 experiment."""

    waits = np.arange(
        params.delay_between_pulses_start,
        params.delay_between_pulses_end,
        params.delay_between_pulses_step,
    )

    sequence, delays = ramsey_sequence(platform, targets)

    data = T2Data()

    sweeper = Sweeper(
        parameter=Parameter.duration,
        values=waits,
        pulses=delays,
    )

    results = platform.execute(
        [sequence],
        [[sweeper]],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.DISCRIMINATION,
        averaging_mode=AveragingMode.SINGLESHOT,
    )

    for qubit in targets:
        ro_pulse = list(sequence.channel(platform.qubits[qubit].acquisition))[-1]
        probs = probability(results[ro_pulse.id], state=1)
        errors = np.sqrt(probs * (1 - probs) / params.nshots)
        data.register_qubit(
            t1.CoherenceProbType, (qubit), dict(wait=waits, prob=probs, error=errors)
        )
    return data


def _fit(data: T2Data) -> T2Results:
    """The used model is

    .. math::

        y = p_0 - p_1 e^{-x p_2}.
    """
    t2s, fitted_parameters, pcovs, chi2 = utils.exponential_fit_probability(data)
    return T2Results(t2s, fitted_parameters, pcovs, chi2)


t2 = Routine(_acquisition, _fit, utils.plot, t2_signal._update)
"""T2 Routine object."""



"""Adds global CLI options."""

import getpass
import pathlib

import click
import yaml

from ..auto.runcard import Runcard
from .acquisition import acquire as acquisition
from .compare import compare_reports
from .fit import fit as fitting
from .report import report as reporting
from .run import protocols_execution
from .update import update as updating
from .upload import upload_report

__all__ = ["command"]

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group()
def command():
    """Welcome to Qibocal!

    Qibo module to calibrate and characterize self-hosted QPUs.
    """


@command.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    "runcard", metavar="RUNCARD", type=click.Path(exists=True, path_type=pathlib.Path)
)
@click.option(
    "folder",
    "-o",
    type=click.Path(path_type=pathlib.Path),
    help="Output folder. If not provided a standard name will generated.",
)
@click.option(
    "force",
    "-f",
    is_flag=True,
    help="Use --force option to overwrite the output folder.",
)
@click.option(
    "--update/--no-update",
    default=True,
    help="Use --no-update option to avoid updating iteratively the platform."
    "With this option the new runcard will not be produced.",
)
@click.option(
    "--platform",
    default=None,
    help="Name of the Qibolab platform.",
)
@click.option(
    "--backend",
    default=None,
    help="Name of the Qibo backend.,",
)
def run(runcard, folder, force, update, platform, backend):
    """Execute the qubit calibration.

    Arguments:

     - RUNCARD: runcard with declarative inputs.
    """
    runcard = Runcard.load(yaml.safe_load(runcard.read_text(encoding="utf-8")))

    if platform is not None:
        runcard.platform = platform
    if backend is not None:
        runcard.backend = backend

    protocols_execution(runcard, folder, force, update)


@command.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    "runcard", metavar="RUNCARD", type=click.Path(exists=True, path_type=pathlib.Path)
)
@click.option(
    "--folder",
    "-o",
    type=click.Path(path_type=pathlib.Path),
    help="Output folder. If not provided a standard name will generated.",
)
@click.option(
    "force",
    "-f",
    is_flag=True,
    help="Use --force option to overwrite the output folder.",
)
@click.option(
    "--platform",
    default=None,
    help="Name of the Qibolab platform.",
)
@click.option(
    "--backend",
    default=None,
    help="Name of the Qibo backend.,",
)
def acquire(runcard, folder, force, platform, backend):
    """Data acquisition.

    Arguments:

     - RUNCARD: runcard with declarative inputs.
    """
    runcard = Runcard.load(yaml.safe_load(runcard.read_text(encoding="utf-8")))

    if platform is not None:
        runcard.platform = platform
    if backend is not None:
        runcard.backend = backend

    acquisition(runcard, folder, force)


@command.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    "folder", metavar="folder", type=click.Path(exists=True, path_type=pathlib.Path)
)
def update(folder):
    """Update platform configuration.

    All configuration files related to platform will be copied
    in the corresponding QIBOLAB_PLAFORMS folder.

    Arguments:
        - folder: Qibocal output folder.

    """
    updating(folder)


@command.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    "folder", metavar="folder", type=click.Path(exists=True, path_type=pathlib.Path)
)
def report(folder):
    """Report generation.

    Arguments:

    - FOLDER: input folder.
    """
    reporting(folder)


@command.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    "input_folder",
    metavar="input_folder",
    type=click.Path(exists=True, path_type=pathlib.Path),
)
@click.option(
    "output_folder",
    "-o",
    type=click.Path(path_type=pathlib.Path),
    help="Output folder where fit is generated.",
)
@click.option(
    "force",
    "-f",
    is_flag=True,
    help="Use --force option to overwrite the output folder.",
)
@click.option(
    "--update/--no-update",
    default=True,
    help="Use --no-update option to avoid updating iteratively the platform."
    "With this option the new runcard will not be produced.",
)
def fit(
    input_folder: pathlib.Path, update: bool, output_folder: pathlib.Path, force: bool
):
    """Post-processing analysis.

    Arguments:

    - FOLDER: input folder.
    """
    fitting(input_folder, update, output_folder, force)


@command.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    "path", metavar="FOLDER", type=click.Path(exists=True, path_type=pathlib.Path)
)
@click.option(
    "--tag",
    default=None,
    type=str,
    help="Optional tag.",
)
@click.option(
    "--author",
    default=getpass.getuser(),
    type=str,
    help="Default is UID username.",
)
def upload(path, tag, author):
    """Uploads output folder to server.

    Arguments:

    - FOLDER: input folder.
    """
    upload_report(path, tag, author)


@command.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    "report_1_path",
    metavar="RUNCARD_1_PATH",
    type=click.Path(exists=True, path_type=pathlib.Path),
)
@click.argument(
    "report_2_path",
    metavar="RUNCARD_2_PATH",
    type=click.Path(exists=True, path_type=pathlib.Path),
)
@click.option(
    "folder",
    "-o",
    type=click.Path(path_type=pathlib.Path),
    help="Output folder. If not provided a standard name will generated.",
)
@click.option(
    "force",
    "-f",
    is_flag=True,
    help="Use --force option to overwrite the output folder.",
)
def compare(report_1_path, report_2_path, folder, force):
    compare_reports(folder, report_1_path, report_2_path, force)


@command.command(context_settings=CONTEXT_SETTINGS, deprecated=True)
@click.argument(
    "runcard", metavar="RUNCARD", type=click.Path(exists=True, path_type=pathlib.Path)
)
@click.option(
    "folder",
    "-o",
    type=click.Path(path_type=pathlib.Path),
    help="Output folder. If not provided a standard name will generated.",
)
@click.option(
    "force",
    "-f",
    is_flag=True,
    help="Use --force option to overwrite the output folder.",
)
@click.option(
    "--update/--no-update",
    default=True,
    help="Use --no-update option to avoid updating iteratively the platform."
    "With this option the new runcard will not be produced.",
)
@click.option(
    "--platform",
    default=None,
    help="Name of the Qibolab platform.",
)
@click.option(
    "--backend",
    default=None,
    help="Name of the Qibo backend.,",
)
def auto(runcard, folder, force, update, platform, backend):
    """Execute the qubit calibration.

    Arguments:

     - RUNCARD: runcard with declarative inputs.
    """
    click.echo(
        "Warning: This command is deprecated and may be removed in a future version. Please use 'qq run' instead. ",
        err=True,
    )



import io
from dataclasses import dataclass, field
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from qibocal.auto.history import History
from qibocal.auto.operation import QubitId, QubitPairId
from qibocal.auto.output import Output
from qibocal.auto.task import Completed, TaskId
from qibocal.cli.report import generate_figures_and_report


@dataclass
class ComparedReport:
    """Class for comparison of two qibocal reports."""

    report_paths: list[Path]
    """List of paths of qibocal reports to compare."""
    folder: Path
    """Generated report path."""
    meta: dict
    """Meta data."""
    history: dict = field(default_factory=dict)
    """History of protocols with same id in both reports."""

    def __post_init__(self):
        """Store only protocols with same id in all the reports."""
        self.history = self.create_common_history()

    def history_uids(self) -> set[TaskId]:
        """Find the set of TakId in common between the reports."""
        experiment_ids = []
        for path in self.report_paths:
            report = Output.load(path)
            experiment_ids.append({x for x in report.history})

        common_experiment_ids = experiment_ids[0]
        for exp_id in experiment_ids[1:]:
            common_experiment_ids = common_experiment_ids & exp_id
        return common_experiment_ids

    def create_common_history(self) -> dict[TaskId, list[History]]:
        """Obtain histories of common TaskIds from the reports."""
        common_experiment_ids = self.history_uids()
        history = {}
        # create a dictionary of task_uid: List[Completed(...)]
        for common_experiment_id in common_experiment_ids:
            history[common_experiment_id] = []
            for path in self.report_paths:
                report = Output.load(path)
                history[common_experiment_id].append(
                    report.history[common_experiment_id]
                )
        return history

    @staticmethod
    def routine_name(routine: TaskId) -> str:
        """Prettify routine's name for report headers."""
        return routine.id.title()

    def routine_targets(self, task_id: TaskId):
        """Extract local targets parameter from Task.

        If not available use the global ones.
        """
        local_targets = self.history[task_id][0].task.targets
        return local_targets if len(local_targets) > 0 else self.meta["targets"]

    def merge_plots(self, plots: list[list[go.Figure]]) -> list[go.Figure]:
        """Merge plots from different reports.

        Scatter plots are plotted in the same figure.
        Heatmaps are vertically stacked.
        """
        merged_figures = []
        for fig0, fig1 in zip(plots[0], plots[1]):
            for trace_0, trace_1 in zip(fig0.data, fig1.data):
                trace_0.legendgroup = None
                trace_1.legendgroup = None
                trace_0.legend = "legend"
                trace_1.legend = "legend2"

            if any(isinstance(trace, go.Heatmap) for trace in fig0.data):
                # TODO: check if this is valid for all protocols
                fig = make_subplots(rows=2, cols=2)
                for j, trace in enumerate(fig0.data):
                    # TODO: remove harcoded
                    # usually with heatmap first should be signal col = 1
                    # then phase (col = 2)
                    # if there is another element it should be a fit done on the first column
                    if isinstance(trace, go.Heatmap):
                        flatten_trace = np.array(trace.z).flatten()
                        flatten_fig_data = np.array(fig1.data[j].z).flatten()
                        trace.zmin = min(min(flatten_trace), min(flatten_fig_data))
                        trace.zmax = max(max(flatten_trace), max(flatten_fig_data))
                    fig.add_trace(
                        trace,
                        row=1,
                        col=1 if j % 2 == 0 else 2,
                    )

                for j, trace in enumerate(fig1.data):
                    # same as before but for second report everything goes
                    # on second row
                    if isinstance(trace, go.Heatmap):
                        trace.showscale = False
                        flatten_trace = np.array(trace.z).flatten()
                        flatten_fig_data = np.array(fig1.data[j].z).flatten()
                        trace.zmin = min(min(flatten_trace), min(flatten_fig_data))
                        trace.zmax = max(max(flatten_trace), max(flatten_fig_data))
                    fig.add_trace(
                        trace,
                        row=2,
                        col=1 if j % 2 == 0 else 2,
                    )
                fig.update_layout(
                    dict(
                        legend={
                            "title": f"{self.report_paths[0].name}",
                            "xref": "container",
                            "yref": "container",
                            "y": 0.8,
                        },
                        legend2={
                            "title": f"{self.report_paths[1].name}",
                            "xref": "container",
                            "yref": "container",
                            "y": 0.2,
                        },
                        showlegend=None,
                    )
                )
                merged_figures.append(fig)

            else:
                for trace in fig1.data:
                    fig0.add_trace(trace)

                fig0.update_layout(
                    dict(
                        legend={
                            "title": f"{self.report_paths[0].name}",
                            "xref": "container",
                            "yref": "container",
                            "y": 0.8,
                        },
                        legend2={
                            "title": f"{self.report_paths[1].name}",
                            "xref": "container",
                            "yref": "container",
                            "y": 0.2,
                        },
                        showlegend=None,
                    )
                )

                # this fixes weird behavior for comparing classification protocols
                if any(isinstance(trace, go.Contour) for trace in fig0.data):
                    fig0.update_layout(
                        dict(
                            legend={
                                "font": {"size": None},
                                "x": None,
                                "xanchor": None,
                                "yanchor": None,
                                "orientation": None,
                            },
                        )
                    )
                    for trace in fig0.data:
                        if isinstance(trace, go.Scatter):
                            if trace.legend == "legend":
                                if trace.name == "Qubit Fit: state 0":
                                    trace.legendgroup = "legend0_q0"
                                else:
                                    trace.legendgroup = "legend0_q1"
                            else:
                                if trace.name == "Qubit Fit: state 0":
                                    trace.legendgroup = "legend1_q0"
                                else:
                                    trace.legendgroup = "legend1_q1"
                merged_figures.append(fig0)

        buffer = io.StringIO()
        html_list = []
        for figure in merged_figures:
            figure.write_html(buffer, include_plotlyjs=True, full_html=False)
            buffer.seek(0)
            html_list.append(buffer.read())
        buffer.close()
        return "".join(html_list)

    def plotter(
        self, nodes: list[Completed], target: Union[QubitId, QubitPairId, list[QubitId]]
    ) -> tuple[str, str]:
        tables = []
        plots = []
        for node in nodes:
            report_figures, fitting_report = generate_figures_and_report(node, target)
            tables.append(fitting_report)
            plots.append(report_figures)

        figures_html = self.merge_plots(plots)

        try:
            merged_table = None
            merge_columns = {"Qubit", "Parameters"}
            for i, table in enumerate(tables):
                if len(table) > 0:
                    a = pd.read_html(io.StringIO(table))[0]
                    a = a.rename(
                        columns={
                            col: f"{col}\n{self.report_paths[i].name}"
                            for col in set(a.columns) - merge_columns
                        }
                    )
                    if merged_table is None:
                        merged_table = a
                    else:
                        merged_table = pd.merge(merged_table, a, on=list(merge_columns))
        except ValueError:
            merged_table = pd.DataFrame(
                [], columns=["An error occurred when performing the fit."]
            )
        if merged_table is not None:
            fitting_report = merged_table.to_html(
                classes="fitting-table", index=False, border=0, escape=False
            )
            return figures_html, fitting_report
        return figures_html, ""



"""SWAP experiment for two qubit gates, chevron plot."""

from dataclasses import dataclass, field

import numpy as np
import numpy.typing as npt
from qibolab import AcquisitionType, AveragingMode, Parameter, Sweeper

from qibocal.auto.operation import QubitPairId, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.result import magnitude

from ...utils import readout_frequency
from ..utils import order_pair
from .chevron import (
    ChevronData,
    ChevronParameters,
    ChevronResults,
    _fit,
    _plot,
    _update,
)
from .utils import chevron_sequence

__all__ = ["chevron_signal"]


@dataclass
class ChevronSignalParameters(ChevronParameters):
    """ChevronSignal runcard inputs."""


@dataclass
class ChevronSignalResults(ChevronResults):
    """ChevronSignal outputs when fitting will be done."""


ChevronSignalType = np.dtype(
    [
        ("amp", np.float64),
        ("length", np.float64),
        ("signal_high", np.float64),
        ("signal_low", np.float64),
    ]
)
"""Custom dtype for Chevron."""


@dataclass
class ChevronSignalData(ChevronData):
    """Chevron acquisition outputs."""

    data: dict[QubitPairId, npt.NDArray[ChevronSignalType]] = field(
        default_factory=dict
    )

    def register_qubit(
        self, low_qubit, high_qubit, length, amp, signal_low, signal_high
    ):
        """Store output for single qubit."""
        size = len(length) * len(amp)
        amplitude, duration = np.meshgrid(amp, length)
        ar = np.empty(size, dtype=ChevronSignalType)
        ar["length"] = duration.ravel()
        ar["amp"] = amplitude.ravel()
        ar["signal_low"] = signal_low.ravel()
        ar["signal_high"] = signal_high.ravel()
        self.data[low_qubit, high_qubit] = np.rec.array(ar)

    def low_frequency(self, pair):
        return self[pair].signal_low

    def high_frequency(self, pair):
        return self[pair].signal_high


def _aquisition(
    params: ChevronSignalParameters,
    platform: CalibrationPlatform,
    targets: list[QubitPairId],
) -> ChevronSignalData:
    r"""
    Perform an iSWAP/CZ experiment between pairs of qubits by changing its frequency.

    Args:
        params: Experiment parameters.
        platform: CalibrationPlatform to use.
        targets (list): List of pairs to use sequentially.

    Returns:
        ChevronData: Acquisition data.
    """

    # create a DataUnits object to store the results,
    data = ChevronSignalData(native=params.native)
    for pair in targets:
        # order the qubits so that the low frequency one is the first
        ordered_pair = order_pair(pair, platform)
        sequence, flux_pulse, parking_pulses, delays = chevron_sequence(
            platform=platform,
            ordered_pair=ordered_pair,
            duration_max=params.duration_max,
            parking=params.parking,
            dt=params.dt,
            native=params.native,
        )

        sweeper_amplitude = Sweeper(
            parameter=Parameter.amplitude,
            range=(params.amplitude_min, params.amplitude_max, params.amplitude_step),
            pulses=[flux_pulse],
        )
        sweeper_duration = Sweeper(
            parameter=Parameter.duration,
            range=(params.duration_min, params.duration_max, params.duration_step),
            pulses=[flux_pulse] + delays + parking_pulses,
        )

        ro_high = list(sequence.channel(platform.qubits[ordered_pair[1]].acquisition))[
            -1
        ]
        ro_low = list(sequence.channel(platform.qubits[ordered_pair[0]].acquisition))[
            -1
        ]

        data.native_amplitude[ordered_pair] = flux_pulse.amplitude

        results = platform.execute(
            [sequence],
            [[sweeper_duration], [sweeper_amplitude]],
            updates=[
                {
                    platform.qubits[q].probe: {
                        "frequency": readout_frequency(q, platform)
                    }
                }
                for q in pair
            ],
            nshots=params.nshots,
            relaxation_time=params.relaxation_time,
            acquisition_type=AcquisitionType.INTEGRATION,
            averaging_mode=AveragingMode.CYCLIC,
        )
        data.register_qubit(
            ordered_pair[0],
            ordered_pair[1],
            sweeper_duration.values,
            sweeper_amplitude.values,
            magnitude(results[ro_low.id]),
            magnitude(results[ro_high.id]),
        )
    return data


chevron_signal = Routine(_aquisition, _fit, _plot, _update, two_qubit_gates=True)
"""Chevron routine."""
